/* GIMP Plugin Preview Widget                                                 
 * Copyright (C) 1998-1999 Shawn T. Amundson                
 * Modifications (C) 1999-2003 Ernst Lippe
 * 
 * Some of the checkerboard & image combination code is based of
 * code Copyright 1997 (C) Federico Mena Quintero and is used
 * here by permission.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.           
 *                                                                            
 * This library is distributed in the hope that it will be useful,             
 * but WITHOUT ANY WARRANTY; without even the implied warranty of              
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Version $Id: gimppreview.c,v 1.10 2003/06/11 11:09:07 el Exp $
 */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <gtk/gtk.h>
#include <glib-object.h>
#include "gimppreview.h"
#include "gimppreviewimage.h"

#ifndef lint
static char vcid[] G_GNUC_UNUSED =
  "$Id: gimppreview.c,v 1.10 2003/06/11 11:09:07 el Exp $";
#endif /* lint */

#define PREVIEW_DEFAULT_SIZE         128
#define PROGRESS_BAR_HEIGHT (10)

/* Minimum time in microseconds between succesive calls to
   the GLib main loop by gimp_preview_process_events.
*/
#define MAIN_LOOP_WAIT_TIME 100000

static void gimp_preview_init (GimpPreview * preview);
static void gimp_preview_class_init (GimpPreviewClass * klass);
static void gimp_preview_destroy (GtkObject * object);

static void gimp_preview_plus_callback (GtkWidget * widget, gpointer data);
static void gimp_preview_minus_callback (GtkWidget * widget, gpointer data);
static gint gimp_preview_event (GtkWidget * widget, GdkEvent * event,
                                gpointer data);
static void gimp_preview_realized_signal_handler (GtkWidget * widget,
                                                  gpointer user_data);
static void gimp_preview_update_handler_internal (GtkWidget * widget,
                                                  GimpPreviewImageUpdateEvent
                                                  * event, gpointer data);
static void scale_changed_handler (GtkAdjustment * adjustment, gpointer data);
static void gimp_preview_set_scale_nr (GimpPreview * preview, gint scale_nr);
static gint gimp_preview_find_scale_nr (GimpPreview * preview, gdouble scale);
static void gimp_preview_set_scales_internal (GimpPreview * preview,
                                              gint n_scales,
                                              const gdouble * scales);



enum
{
  UPDATE_PREVIEW,
  LAST_SIGNAL
};

#define N_DEFAULT_SCALES 21
static const gdouble default_scales[N_DEFAULT_SCALES] = {
  1 / 6.0, 1 / 5.0, 1 / 4.0,
  1 / 3.0, 1 / 2.0, 1.00,
  2.00, 3.00, 4.00,
  5.00, 6.00, 7.00,
  8.00, 9.00, 10.00,
  11.00, 12.00, 13.00,
  14.00, 15.00, 16.00
};

static guint gimp_preview_signals[LAST_SIGNAL] = { 0 };

GType
gimp_preview_get_type (void)
{
  static GType preview_type = 0;

  if (!preview_type)
    {
      GTypeInfo preview_info = {
        sizeof (GimpPreviewClass),
        (GBaseInitFunc) NULL,
        (GBaseFinalizeFunc) NULL,
        (GClassInitFunc) gimp_preview_class_init,
        (GClassFinalizeFunc) NULL,
        (gconstpointer) NULL,   /* class_data */
        sizeof (GimpPreview),
        0,                      /* n_preallocs */
        (GInstanceInitFunc) gimp_preview_init,
        (GTypeValueTable *) NULL        /* value_table */
      };

      preview_type =
        g_type_register_static (GTK_TYPE_TABLE, "GimpPreview",
                                &preview_info, 0);
    }

  return preview_type;
}


/*
 * Initialization, which is done when the preview widget is first created
 * by GTK's internal mechanisms.
 */
static void
gimp_preview_class_init (GimpPreviewClass * klass)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);

  gimp_preview_signals[UPDATE_PREVIEW] =
    g_signal_new ("update-preview",
                  G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_FIRST,
                  G_STRUCT_OFFSET (GimpPreviewClass, update_preview),
                  NULL,
                  NULL,
                  g_cclosure_marshal_VOID__POINTER,
                  G_TYPE_NONE, 1, G_TYPE_POINTER);

  klass->update_preview = NULL;
  object_class->destroy = gimp_preview_destroy;
}

/*
 * Initialization, which is done when the preview widget is first created
 * by GTK's internal mechanisms.
 */
static void
gimp_preview_init (GimpPreview * preview)
{
  g_assert (GIMP_IS_PREVIEW (preview));

  preview->scales = NULL;
  gimp_preview_set_scales_internal (preview, N_DEFAULT_SCALES,
                                    default_scales);
  preview->in_drag = FALSE;
  preview->drag_x = 0;
  preview->drag_y = 0;
  preview->orig_preview_x = 0;
  preview->orig_preview_y = 0;
  preview->previous_event_time.tv_sec = 0;
  preview->previous_event_time.tv_usec = 0;

  preview->label = NULL;
  preview->button_minus = NULL;
  preview->button_plus = NULL;
  preview->zoom_controls = NULL;
  preview->event_box = NULL;
  preview->image = NULL;
  preview->progress_bar = NULL;
  preview->hscroll_bar = NULL;
  preview->vscroll_bar = NULL;
}

static void
gimp_preview_destroy (GtkObject * object)
{
  GimpPreview *preview = GIMP_PREVIEW (object);
  GtkObjectClass *parent_class =
    GTK_OBJECT_CLASS (g_type_class_peek_parent
                      (G_OBJECT (preview)->g_type_instance.g_class));

  g_free (preview->scales);
  preview->scales = NULL;

  parent_class->destroy (object);
}


/**
 * gimp_preview_new:
 * @drawable: The original drawable that will be shown when the user
 * is scrolling.
 * 
 * Create a new #GimpPreview widget.
 * This call creates a widget with a default size,
 * initial scale = 1 and with a visible progress-bar 
 * zoom controls, and invisible scroll-bars.
 * 
 * Return value: The new GimpPreview widget. 
 **/
GtkWidget *
gimp_preview_new (GimpDrawable * drawable)
{
  /* Simplified front end that uses all the defaults. */
  return (gimp_preview_new_with_args
          (drawable, -1, -1, 1.0, TRUE, TRUE, FALSE));
}

/**
 * gimp_preview_new_with_args:
 * @drawable: The original drawable that will be shown when the user
 * is scrolling.
 * @preview_width: Width of the preview image in pixels. When it is
 * < 0 a default width is used.
 * @preview_height: Heigth of the preview image in pixels. When it is
 * < 0 a default heigth is used.
 * @initial_scale: Initial scale of the preview.
 * @show_progressbar: Is the progress-bar visible.
 * @show_zoom_controls: Are the zoom controls visible.
 * @show_scroll_bars: Are scroll-bars visible.
 * 
 * Create a new #GimpPreview widget.
 * 
 * Return value: The new GimpPreview widget. 
 **/
GtkWidget *
gimp_preview_new_with_args (GimpDrawable * drawable,
                            gint preview_width,
                            gint preview_height,
                            gdouble initial_scale,
                            gboolean show_progressbar,
                            gboolean show_zoom_controls,
                            gboolean show_scroll_bars)
{
  GimpPreview *preview;
  GtkWidget *zoom_hbox;
  GtkWidget *event_box;
  GtkWidget *zoom_in_image, *zoom_out_image;

  if (preview_width < 0)
    {
      preview_width = PREVIEW_DEFAULT_SIZE;
    };
  if (preview_height < 0)
    {
      preview_height = PREVIEW_DEFAULT_SIZE;
    };

  /* Now allocate the actual preview window. */
  preview = GIMP_PREVIEW (g_object_new (gimp_preview_get_type (), NULL));
  gtk_table_resize (GTK_TABLE (preview), 4, 2);

  /* Now build our dialog. */

  event_box = gtk_event_box_new ();
  preview->image =
    GIMP_PREVIEW_IMAGE (gimp_preview_image_new_with_args
                        (drawable, preview_width, preview_height,
                         initial_scale));
  gtk_container_add (GTK_CONTAINER (event_box), GTK_WIDGET (preview->image));
  preview->event_box = event_box;
  gtk_table_attach (GTK_TABLE (preview), event_box, 0, 1, 0, 1,
                    GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                    GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
  g_signal_connect_after (G_OBJECT (preview->image),
                          "realize",
                          G_CALLBACK (gimp_preview_realized_signal_handler),
                          preview);

  g_signal_connect (G_OBJECT (preview->image),
                    "update-preview",
                    G_CALLBACK (gimp_preview_update_handler_internal),
                    preview);

  gtk_widget_set_events (event_box,
                         GDK_BUTTON_PRESS_MASK |
                         GDK_BUTTON_RELEASE_MASK |
                         GDK_POINTER_MOTION_HINT_MASK |
                         GDK_BUTTON_MOTION_MASK);
  g_signal_connect (G_OBJECT (event_box), "event",
                    G_CALLBACK (gimp_preview_event), (gpointer) preview);

  preview->hscroll_bar =
    gtk_hscrollbar_new (gimp_preview_get_x_adjustment (preview));
  gtk_table_attach (GTK_TABLE (preview), preview->hscroll_bar,
                    0, 1, 1, 2, GTK_SHRINK | GTK_FILL, 0, 0, 0);
  preview->vscroll_bar =
    gtk_vscrollbar_new (gimp_preview_get_y_adjustment (preview));
  gtk_table_attach (GTK_TABLE (preview), preview->vscroll_bar,
                    1, 2, 0, 1, 0, GTK_SHRINK | GTK_FILL, 0, 0);
  if (show_scroll_bars)
    {
      gtk_widget_show (preview->hscroll_bar);
      gtk_widget_show (preview->vscroll_bar);
    }


  /* Progress-bar */
  preview->progress_bar = gtk_progress_bar_new ();
  gtk_widget_set_size_request (preview->progress_bar, -1,
                               PROGRESS_BAR_HEIGHT);
  gtk_table_attach (GTK_TABLE (preview), preview->progress_bar,
                    0, 2, 2, 3, GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0, 0);

  if (show_progressbar)
    {
      gtk_widget_show (preview->progress_bar);
    }

  zoom_hbox = gtk_hbox_new (FALSE, 0);
  preview->zoom_controls = zoom_hbox;
  gtk_container_set_resize_mode (GTK_CONTAINER (zoom_hbox),
                                 GTK_RESIZE_IMMEDIATE);
  gtk_container_set_border_width (GTK_CONTAINER (zoom_hbox), 4);

  preview->button_minus = gtk_button_new ();
  zoom_out_image =
    gtk_image_new_from_stock (GTK_STOCK_ZOOM_OUT, GTK_ICON_SIZE_BUTTON);
  gtk_container_add (GTK_CONTAINER (preview->button_minus), zoom_out_image);
  preview->button_plus = gtk_button_new ();
  zoom_in_image =
    gtk_image_new_from_stock (GTK_STOCK_ZOOM_IN, GTK_ICON_SIZE_BUTTON);
  gtk_container_add (GTK_CONTAINER (preview->button_plus), zoom_in_image);

  preview->label = gtk_label_new ("100%");
  g_signal_connect (G_OBJECT (gimp_preview_image_get_scale_adjustment
                              (preview->image)),
                    "value-changed",
                    G_CALLBACK (scale_changed_handler), preview);

  gtk_box_pack_start (GTK_BOX (zoom_hbox),
                      preview->button_minus, TRUE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX (zoom_hbox), preview->label, TRUE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX (zoom_hbox),
                      preview->button_plus, TRUE, FALSE, 0);

  g_signal_connect (G_OBJECT (preview->button_minus),
                    "clicked",
                    G_CALLBACK (gimp_preview_minus_callback),
                    (gpointer) preview);
  g_signal_connect (G_OBJECT (preview->button_plus),
                    "clicked",
                    G_CALLBACK (gimp_preview_plus_callback),
                    (gpointer) preview);

  gtk_widget_show (zoom_in_image);
  gtk_widget_show (zoom_out_image);
  gtk_widget_show (preview->button_minus);
  gtk_widget_show (preview->label);
  gtk_widget_show (preview->button_plus);
  gtk_table_attach (GTK_TABLE (preview), zoom_hbox, 0, 2, 3, 4,
                    GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0, 0);

  /* Set the scale amount. */
  gimp_preview_set_scale_nr (preview,
                             gimp_preview_find_scale_nr (preview,
                                                         initial_scale));

  if (show_zoom_controls)
    {
      gtk_widget_show (zoom_hbox);
    }

  gtk_widget_show (GTK_WIDGET (preview->image));
  gtk_widget_show (event_box);

  /*
   * Build our initial preview of the drawable data.
   */

  gimp_preview_image_update (preview->image);

  return GTK_WIDGET (preview);
}

/**
 * gimp_preview_update:
 * @preview: A #GimpPreview.
 * 
 * Plug-ins call this to do an update of the preview area.
 * The preview will generate an "#update-preview" signal.
 * The plug-in's signal handler should then draw in on the preview
 * with one of
 * the drawing functions: gimp_preview_draw_unscaled_row(),
 * gimp_preview_draw_unscaled_drawable(), gimp_preview_draw_scaled_row()
 * or gimp_preview_draw_scaled_drawable().
 **/
void
gimp_preview_update (GimpPreview * preview)
{
  gimp_preview_image_update (preview->image);
}

/* Find highest scale_nr with value <= scale  */
static gint
gimp_preview_find_scale_nr (GimpPreview * preview, gdouble scale)
{
  gint scale_nr = preview->n_scales - 1;

  while ((scale_nr > 0) && (preview->scales[scale_nr] > scale))
    {
      scale_nr--;
    }
  return scale_nr;
}

static void
gimp_preview_set_scale_nr (GimpPreview * preview, gint scale_nr)
{
  scale_nr = CLAMP (scale_nr, 0, preview->n_scales - 1);

  gimp_preview_image_set_scale (preview->image, preview->scales[scale_nr]);
}

/* This handler is called when the scale changes.
 * The contents of the scale label is updated.
 */
static void
scale_changed_handler (GtkAdjustment * adjustment, gpointer data)
{
  GimpPreview *preview = GIMP_PREVIEW (data);
  const gint buffer_size = 10;
  gchar buffer[buffer_size];
  gdouble new_scale = adjustment->value;

  g_snprintf (buffer, buffer_size, "%d%%", (gint) (new_scale * 100));
  gtk_label_set_text (GTK_LABEL (preview->label), buffer);

  gtk_widget_set_sensitive (preview->button_plus,
                            new_scale <
                            preview->scales[preview->n_scales - 1]);
  gtk_widget_set_sensitive (preview->button_minus,
                            new_scale > preview->scales[0]);
}

/*
 * Change the scale amount upwards and redraw the display.
 */
static void
gimp_preview_plus_callback (GtkWidget * widget, gpointer data)
{
  GimpPreview *preview = GIMP_PREVIEW (data);
  const gdouble scale = gimp_preview_get_scale (preview);
  const gint scale_nr = gimp_preview_find_scale_nr (preview, scale);

  if (scale < preview->scales[scale_nr])
    {
      gimp_preview_set_scale_nr (preview, scale_nr);
    }
  else
    {
      gimp_preview_set_scale_nr (preview, scale_nr + 1);
    }
}


/*
 * Change the scale amount downwards and redraw the display.
 */
static void
gimp_preview_minus_callback (GtkWidget * widget, gpointer data)
{
  GimpPreview *preview = GIMP_PREVIEW (data);

  const gdouble scale = gimp_preview_get_scale (preview);
  const gint scale_nr = gimp_preview_find_scale_nr (preview, scale);

  if (scale > preview->scales[scale_nr])
    {
      gimp_preview_set_scale_nr (preview, scale_nr);
    }
  else
    {
      gimp_preview_set_scale_nr (preview, scale_nr - 1);
    }
}


/*
 * This function handles events from the mouse - dragging, button presses
 * and releases.  We only notice button 1 on the mouse to initiate and manage
 * dragging of the preview.
 */
static gint
gimp_preview_event (GtkWidget * widget, GdkEvent * event, gpointer data)
{
  GimpPreview *preview;
  GdkEventButton *button_event;
  gint x, y;
  gint dx, dy;
  gint dummy1, dummy2;

  preview = GIMP_PREVIEW (data);
  button_event = (GdkEventButton *) event;

  switch (event->type)
    {
    case GDK_BUTTON_PRESS:
      if (button_event->button == 1)
        {
          gtk_widget_get_pointer (widget, &x, &y);

          preview->in_drag = TRUE;
          preview->drag_x = x;
          preview->drag_y = y;

          gimp_preview_image_get_preview_position (GIMP_PREVIEW_IMAGE
                                                   (preview->
                                                    image),
                                                   &(preview->
                                                     orig_preview_x),
                                                   &(preview->
                                                     orig_preview_y), &dummy1,
                                                   &dummy2);

          gtk_grab_add (widget);
        }
      break;

    case GDK_BUTTON_RELEASE:
      if (preview->in_drag && button_event->button == 1)
        {
          gtk_grab_remove (widget);
          preview->in_drag = FALSE;
        }
      break;

    case GDK_MOTION_NOTIFY:
      if (preview->in_drag)
        {
          gtk_widget_get_pointer (widget, &x, &y);

          dx = x - preview->drag_x;
          dy = y - preview->drag_y;

          gimp_preview_image_set_position (preview->image,
                                           preview->
                                           orig_preview_x - dx,
                                           preview->orig_preview_y - dy);
        }
      break;

    default:
      break;
    }

  return FALSE;
}


/**
 * gimp_preview_draw_scaled_drawable:
 * @preview: a #GimpPreview.
 * @drawable: A #GimpDrawable that contains the scaled rendered image.
 * 
 * Draw a scaled drawable in the preview. 
 * 
 * It is only possible to draw from within an "update-preview"
 * signal handler when the user has not scrolled or zoomed
 * after the invocation of this signal handler.
 * In all other cases drawing on the preview is not
 * allowed and all drawing requests are ignored.
 * Rendering functions should halt immediately when
 * this function returns %FALSE, because the results of
 * this rendering function will be ignored by the preview.
 *
 * Return value: %TRUE when the drawing on the preview is allowed.
 * When this function returns %FALSE the rendering should
 * halt immediately.
 **/
gboolean
gimp_preview_draw_scaled_drawable (GimpPreview * preview,
                                   GimpDrawable * drawable)
{
  return gimp_preview_image_draw_scaled_drawable (preview->image, drawable);
}

/**
 * gimp_preview_draw_unscaled_drawable:
 * @preview: a #GimpPreview.
 * @drawable: A #GimpDrawable that contains the unscaled rendered image.
 * 
 * Draw an unscaled drawable in the preview. 
 * 
 * It is only possible to draw from within an "update-preview"
 * signal handler when the user has not scrolled or zoomed
 * after the invocation of this signal handler.
 * In all other cases drawing on the preview is not
 * allowed and all drawing requests are ignored.
 * Rendering functions should halt immediately when
 * this function returns %FALSE, because the results of
 * this rendering function will be ignored by the preview.
 *
 * Return value: %TRUE when the drawing on the preview is allowed.
 * When this function returns %FALSE the rendering should
 * halt immediately.
 **/
gboolean
gimp_preview_draw_unscaled_drawable (GimpPreview * preview,
                                     GimpDrawable * drawable)
{
  return
    gimp_preview_image_draw_unscaled_drawable (GIMP_PREVIEW_IMAGE
                                               (preview->image), drawable);
}

/**
 * gimp_preview_draw_scaled_row:
 * @preview: the #GimpPreview
 * @type: the format of the data (e.g. %GIMP_RGBA_IMAGE).
 * @row:the relative number of the row within the preview.
 * The top row of the preview is number 0.
 * @width: the number of pixels in this row.
 * @data: pixels for the preview. It must have a length of width pixels
 *
 * Draw one scaled row of data in the preview.
 * This function handles conversion to RGB and checkerboarding of
 * transparent areas.
 *
 * It is only possible to draw from within an "update-preview"
 * signal handler when the user has not scrolled or zoomed
 * after the invocation of this signal handler.
 * In all other cases drawing on the preview is not
 * allowed and all drawing requests are ignored.
 * Rendering functions should halt immediately when
 * this function returns %FALSE, because the results of
 * this rendering function will be ignored by the preview.
 *
 * Return value: %TRUE when the drawing on the preview is allowed.
 * When this function returns %FALSE the rendering should
 * halt immediately.
 **/
gboolean
gimp_preview_draw_scaled_row (GimpPreview * preview,
                              GimpImageType type, gint row, gint width,
                              const guchar * data)
{
  return
    gimp_preview_image_draw_scaled_row (preview->image,
                                        type, row, width, data);
}

/**
 * gimp_preview_draw_unscaled_row:
 * @preview: the #GimpPreview
 * @type: the format of the data (e.g. %GIMP_RGBA_IMAGE).
 * @row:row is the relative position of the row w.r.t. preview_event->image_y.
 * The top row has number 0.
 * @width: the number of pixels in this row.
 * @data: pixels for the preview. The first pixel in @data has x-coordinate
 * preview_event->image_x in the image. @data must have a length
 * of @width pixels.
 *
 * Draw one unscaled row of data in the preview.
 * This function handles scaling, conversion to RGB and checkerboarding of
 * transparent areas.
 * A nice feature is that this function will draw several lines of
 * the preview when scale > 1.
 *
 * It is only possible to draw from within an "update-preview"
 * signal handler when the user has not scrolled or zoomed
 * after the invocation of this signal handler.
 * In all other cases drawing on the preview is not
 * allowed and all drawing requests are ignored.
 * Rendering functions should halt immediately when
 * this function returns %FALSE, because the results of
 * this rendering function will be ignored by the preview.
 *
 * Return value: %TRUE when the drawing on the preview is allowed.
 * When this function returns %FALSE the rendering should
 * halt immediately.
 **/
gboolean
gimp_preview_draw_unscaled_row (GimpPreview * preview,
                                GimpImageType type, gint row,
                                gint width, const guchar * data)
{
  return
    gimp_preview_image_draw_unscaled_row (preview->image,
                                          type, row, width, data);
}

/* Return if a > b */
gboolean
g_time_val_greater (const GTimeVal * a, const GTimeVal * b)
{
  return ((a->tv_sec > b->tv_sec) ||
          ((a->tv_sec == b->tv_sec) && (a->tv_usec > b->tv_usec)));
}

/* Call the GLib main loop to process new events */
static void
gimp_preview_process_events (GimpPreview *preview)
{
  GTimeVal now, next_time;

  /* The GLib main loop must be called at regular intervals otherwise
     Gtk waits with updating the display until the plug-in is ready
     with rendering.  This also enable the GUI to process new user
     events like scrolls.

     Because calling the main loop is expensive it is only
     called when at least MAIN_LOOP_WAIT_TIME microseconds have
     passed since the previous call.
   */
  g_get_current_time (&now);
  next_time = preview->previous_event_time;
  g_time_val_add (&next_time, MAIN_LOOP_WAIT_TIME);

  if (g_time_val_greater (&now, &next_time))
    {
      preview->previous_event_time = now;
      while (g_main_iteration (FALSE));
    }
}

/**
 * gimp_preview_progress_set_fraction:
 * @preview: the #GimpPreview.
 * @fraction: the fraction completed.
 *
 * Set the progress bar of the preview to @fraction completed.
 *
 * This function is very useful even when the preview has no visible
 * progress-bar because its return value tells the plug-in whether it
 * should continue its computations.  When the user has scrolled or
 * zoomed, the image that is currently being computed by the rendering
 * function has become obsolete.
 * 
 * When this is the case the rendering function can halt immediately
 * because the preview will not use the result.
 *
 * This function calls the GLib main loop to process new user events.
 * If you don't call this function at regular intervals in
 * your rendering function, the GUI will be "frozen" while
 * the rendering function is busy. This is highly annoying.
 *
 * In most cases you should consider using a #GimpPreviewProgressUpdater
 * instead of calling this function directly, because it makes
 * it easier to write a rendering function that can be used
 * both for the preview and for the final result.
 * 
 * Return value: %TRUE when the drawing on the preview is allowed.
 * When this function returns %FALSE the rendering should
 * halt immediately.
 **/
gboolean
gimp_preview_progress_set_fraction (GimpPreview * preview, double fraction)
{
  gboolean return_status;

  /* The GLib main loop must be called at regular intervals otherwise
     Gtk waits with updating the display until the plug-in is ready
     with rendering.  This also enable the GUI to process new user
     events like scrolls.
   */
  gimp_preview_process_events (preview);
  
  return_status = gimp_preview_image_is_drawing_allowed (preview->image);

  if (return_status)
    {
      gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR
                                     (preview->progress_bar), fraction);
    }
  return (return_status);
}

/**
 * gimp_preview_set_position:
 * @preview: A #GimpPreview.
 * @x_pos: The left side of the preview in #preview-coordinates.
 * @y_pos: The top side of the preview in #preview-coordinates.
 * 
 * Set the position of the area that is show in the preview.
 **/
void
gimp_preview_set_position (GimpPreview * preview, gint x_pos, gint y_pos)
{
  gimp_preview_image_set_position (preview->image, x_pos, y_pos);
}

/**
 * gimp_preview_get_preview_position:
 * @preview: A #GimpPreview. 
 * @prev_x_pos: Left of the preview in #preview-coordinates.
 * @prev_y_pos: Top of the preview in #preview-coordinates.
 * @prev_width: Width of the preview.
 * @prev_height: Height of the preview.
 * 
 * Return the size and position of the preview in #preview-coordinates.
 **/
void
gimp_preview_get_preview_position (GimpPreview * preview,
                                   gint * prev_x_pos, gint * prev_y_pos,
                                   gint * prev_width, gint * prev_height)
{
  gimp_preview_image_get_preview_position (preview->image,
                                           prev_x_pos, prev_y_pos, prev_width,
                                           prev_height);
}

/**
 * gimp_preview_get_image_position:
 * @preview: A #GimpPreview.
 * @im_x_pos: Left of the preview in #image-coordinates.
 * @im_y_pos: Top of the preview in #image-coordinates.
 * @im_width: Width of the preview in #image-coordinates.
 * @im_height: Heigth of the preview in #image-coordinates.
 * 
 * Return the size and position of the preview in #image-coordinates.
 * The preview shows a scaled version of this area of the original or
 * rendered image.
 **/
void
gimp_preview_get_image_position (GimpPreview * preview,
                                 gint * im_x_pos, gint * im_y_pos,
                                 gint * im_width, gint * im_height)
{
  gimp_preview_image_get_image_position (preview->image,
                                         im_x_pos, im_y_pos, im_width,
                                         im_height);
}

/**
 * gimp_preview_set_scale:
 * @preview: A #GimpPreview.
 * @scale: The new scale factor.
 * 
 * Set the current scale factor for the preview.
 **/
void
gimp_preview_set_scale (GimpPreview * preview, gdouble scale)
{
  gimp_preview_image_set_scale (preview->image, scale);
}

/**
 * gimp_preview_get_scale:
 * @preview: A #GimpPreview. 
 * 
 * Get the current scale factor for the preview.
 * 
 * Return value: The scale factor.
 **/
gdouble
gimp_preview_get_scale (GimpPreview * preview)
{
  return gimp_preview_image_get_scale (preview->image);
}

/**
 * gimp_preview_get_progress_bar_visible:
 * @preview: A #GimpPreview.
 * 
 * Get the visibility of the progress_bar
 * 
 * Return value: TRUE when the progress_bar is visible.
 **/
gboolean
gimp_preview_get_progress_bar_visible (GimpPreview * preview)
{
  return (GTK_WIDGET_VISIBLE (preview->progress_bar));
}

/**
 * gimp_preview_set_progress_bar_visible:
 * @preview: A #GimpPreview.
 * @visible: TRUE when the progress_bar must be visible.
 * 
 * Show or hide the progress_bar.
 **/
void
gimp_preview_set_progress_bar_visible (GimpPreview * preview,
                                       gboolean visible)
{
  if (visible)
    {
      gtk_widget_show (preview->progress_bar);
    }
  else
    {
      gtk_widget_hide (preview->progress_bar);
    }
}

/**
 * gimp_preview_get_zoom_controls_visible:
 * @preview: A #GimpPreview. 
 * 
 * 
 * Get the visibility of the zoom controls.
 * 
 * Return value: TRUE when the zoom controls are visible.
 **/
gboolean
gimp_preview_get_zoom_controls_visible (GimpPreview * preview)
{
  return (GTK_WIDGET_VISIBLE (preview->zoom_controls));
}

/**
 * gimp_preview_set_zoom_controls_visible:
 * @preview: A #GimpPreview. 
 * @visible: TRUE when the progress_bar must be visible.
 * 
 * Show or hide the zoom controls.
 **/
void
gimp_preview_set_zoom_controls_visible (GimpPreview * preview,
                                        gboolean visible)
{
  if (visible)
    {
      gtk_widget_show (preview->zoom_controls);
    }
  else
    {
      gtk_widget_hide (preview->zoom_controls);
    }
}

/**
 * gimp_preview_get_scroll_bars_visible:
 * @preview: A #GimpPreview. 
 * 
 * 
 * Get the visibility of the scroll-bars.
 * 
 * Return value: TRUE when the scroll-bars are visible.
 **/
gboolean
gimp_preview_get_scroll_bars_visible (GimpPreview * preview)
{
  return (GTK_WIDGET_VISIBLE (preview->hscroll_bar) ||
          GTK_WIDGET_VISIBLE (preview->vscroll_bar));
}

/**
 * gimp_preview_set_scroll_bars_visible:
 * @preview: A #GimpPreview. 
 * @visible: TRUE when the scroll-bars must be visible.
 * 
 * Show or hide the scroll-bars.
 **/
void
gimp_preview_set_scroll_bars_visible (GimpPreview * preview, gboolean visible)
{
  if (visible)
    {
      gtk_widget_show (preview->hscroll_bar);
      gtk_widget_show (preview->vscroll_bar);
    }
  else
    {
      gtk_widget_hide (preview->hscroll_bar);
      gtk_widget_hide (preview->vscroll_bar);
    }
}

/**
 * gimp_preview_get_show_original:
 * @preview: A #GimpPreview.
 * 
 * Query the @preview whether it shows the original image
 * during scrolls.
 * 
 * Return value: TRUE when the preview shows a scaled version
 * of the original image during scrolls.
 **/
gboolean
gimp_preview_get_show_original (GimpPreview * preview)
{
  g_return_val_if_fail (GIMP_IS_PREVIEW (preview), TRUE);

  return gimp_preview_image_get_show_original (preview->image);
}

/**
 * gimp_preview_set_show_original:
 * @preview: A #GimpPreview.
 * @show_original: TRUE when the preview shows a scaled version
 * of the original image during scrolls.
 * 
 * Set whether the @preview_image should show a scaled version
 * of the original image during scrolls.
 * When show_original is FALSE only the results of the rendering
 * function will be shown in the display.
 * This will only give acceptable results when the rendering
 * function is very fast.
 **/
void
gimp_preview_set_show_original (GimpPreview * preview, gboolean show_original)
{
  g_return_if_fail (GIMP_IS_PREVIEW (preview));

  gimp_preview_image_set_show_original (preview->image, show_original);
}


void gimp_preview_image_set_show_original (GimpPreviewImage * preview_image,
                                           gboolean show_original);

/**
 * gimp_preview_get_x_adjustment:
 * @preview: A #GimpPreview.
 * 
 * Return the adjustment that manages the x-coordinate of
 * the preview position. The value of this adjustment
 * is expressed in #preview-coordinates.
 * 
 * Applications should only change the value of the adjustment,
 * the other properties of the adjustment are managed by the widget.
 * When you try to set the value directly (i.e. not via a GtkScale widget)
 * you should use gtk_adjustment_set_correct_value() (don't use
 * gtk_adjustment_set_value() because it is buggy).
 *
 * Return value: The #GtkAdjustment.
 **/
GtkAdjustment *
gimp_preview_get_x_adjustment (GimpPreview * preview)
{
  g_return_val_if_fail (GIMP_IS_PREVIEW (preview), NULL);
  return gimp_preview_image_get_x_adjustment (preview->image);
}

/**
 * gimp_preview_get_y_adjustment:
 * @preview: A #GimpPreview.
 * 
 * Return the adjustment that manages the y-coordinate of
 * the preview position. The value of this adjustment
 * is expressed in #preview-coordinates.
 * 
 * Applications should only change the value of the adjustment,
 * the other properties of the adjustment are managed by the widget.
 * When you try to set the value directly (i.e. not via a GtkScale widget)
 * you should use gtk_adjustment_set_correct_value() (don't use
 * gtk_adjustment_set_value() because it is buggy).
 *
 * Return value: The #GtkAdjustment.
 **/
GtkAdjustment *
gimp_preview_get_y_adjustment (GimpPreview * preview)
{
  g_return_val_if_fail (GIMP_IS_PREVIEW (preview), NULL);
  return gimp_preview_image_get_y_adjustment (preview->image);
}

/**
 * gimp_preview_get_scale_adjustment:
 * @preview: A #GimpPreview.
 * 
 * Return the adjustment that manages the scale of
 * the preview position.
 * The value of this adjustment
 * is the ratio between the size of the scaled image
 * and the size of the original image, e.g. a value
 * of 0.5 means that the preview image is reduced by
 * 50%.
 * 
 * Applications should only change the value of the adjustment,
 * the other properties of the adjustment are managed by the widget.
 * When you try to set the value directly (i.e. not via a GtkScale widget)
 * you should use gtk_adjustment_set_correct_value() (don't use
 * gtk_adjustment_set_value() because it is buggy).
 *
 * Return value: The #GtkAdjustment.
 **/
GtkAdjustment *
gimp_preview_get_scale_adjustment (GimpPreview * preview)
{
  g_return_val_if_fail (GIMP_IS_PREVIEW (preview), NULL);
  return gimp_preview_image_get_scale_adjustment (preview->image);
}

/* This is the handler for the "update-preview" signal from
 * the GimpPreviewImage. The event is converted and a new
 * "update-preview" signal is emitted by the GimpPreview.
 */
static void
gimp_preview_update_handler_internal (GtkWidget * widget,
                                      GimpPreviewImageUpdateEvent * event,
                                      gpointer data)
{
  GimpPreview *preview = GIMP_PREVIEW (data);
  GimpPreviewUpdateEvent preview_event;

  preview_event.scale = event->scale;
  preview_event.image_x = event->image_x;
  preview_event.image_y = event->image_y;
  preview_event.image_width = event->image_width;
  preview_event.image_height = event->image_height;
  preview_event.preview_x = event->preview_x;
  preview_event.preview_y = event->preview_y;
  preview_event.preview_width = event->preview_width;
  preview_event.preview_height = event->preview_height;

  g_signal_emit (GTK_OBJECT (preview),
                 gimp_preview_signals[UPDATE_PREVIEW], 0, &preview_event);

}

/* Give the preview image a "move" cursor.
 * This can only be done when the image has been realized.
 */
static void
gimp_preview_realized_signal_handler (GtkWidget * widget, gpointer preview)
{
  GtkWidget *preview_image = GTK_WIDGET (GIMP_PREVIEW (preview)->image);

  gdk_window_set_cursor (gtk_widget_get_parent_window (preview_image),
                         gdk_cursor_new (GDK_FLEUR));
}

/* Set a new set of scales.
   Performs some sanity checks.
   */
static void
gimp_preview_set_scales_internal (GimpPreview * preview,
                                  gint n_scales, const gdouble * scales)
{
  gdouble *new_scales = g_new (gdouble, n_scales);
  gint index;
  gdouble previous_scale = -1.0;
  gboolean positive_scales = TRUE;
  gboolean ascending = TRUE;

  for (index = 0; index < n_scales; index++)
    {
      const gdouble scale = scales[index];

      positive_scales &= (scale > 0.0);
      ascending &= (scale > previous_scale);
      new_scales[index] = scale;
      previous_scale = scale;
    }
  g_assert (positive_scales);
  g_assert (ascending);

  preview->n_scales = n_scales;
  g_free (preview->scales);
  preview->scales = new_scales;
}


/**
 * gimp_preview_set_scales:
 * @preview: A #GimpPreview.
 * @n_scales: Number of entries in @scales array.
 * @scales: Array of scale values. All values must be
 * positive and sorted in ascending order.
 * 
 * Set the scales that are used by the zoom buttons.
 **/
void
gimp_preview_set_scales (GimpPreview * preview, gint n_scales,
                         const gdouble * scales)
{
  const gdouble current_scale = gimp_preview_get_scale (preview);
  preview->n_scales = 0;
  preview->scales = NULL;
  gimp_preview_set_scales_internal (preview, n_scales, scales);
  gimp_preview_set_scale_nr (preview,
                             gimp_preview_find_scale_nr (preview,
                                                         current_scale));

}
