/* GIMP Plugin Preview Widget                                                 
 * Copyright (C) 1998-1999 Shawn T. Amundson                
 * Modifications (C) 1999-2003 Ernst Lippe
 * 
 * Some of the checkerboard & image combination code is based of
 * code Copyright 1997 (C) Federico Mena Quintero and is used
 * here by permission.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.           
 *                                                                            
 * This library is distributed in the hope that it will be useful,             
 * but WITHOUT ANY WARRANTY; without even the implied warranty of              
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Version $Id: gimppreviewimage.c,v 1.11 2003/06/11 11:09:10 el Exp $
 */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <gtk/gtk.h>
#include <glib-object.h>
#include "gimppreviewimage.h"

#ifndef lint
static char vcid[] G_GNUC_UNUSED =
  "$Id: gimppreviewimage.c,v 1.11 2003/06/11 11:09:10 el Exp $";
#endif /* lint */

#define PREVIEW_DEFAULT_SIZE         128

static void gimp_preview_image_init (GimpPreviewImage * preview_image);
static void gimp_preview_image_destroy (GtkObject * object);
static void gimp_preview_image_class_init (GimpPreviewImageClass * klass);

static void gimp_preview_image_recompute_sizes (GimpPreviewImage *
                                                preview_image);

static void gimp_preview_image_redraw_original_image (GimpPreviewImage *
                                                      preview_image);
static void draw_unscaled_drawable_with_offset (GimpPreviewImage *
                                                preview_image,
                                                GimpDrawable * drawable,
                                                gint offset_x, gint offset_y);

static void gimp_preview_image_set_size (GimpPreviewImage * preview_image,
                                         gint width, gint height);
static void gimp_preview_image_size_request (GtkWidget * widget,
                                             GtkRequisition * requisition);
static void gimp_preview_image_force_redraw (GimpPreviewImage *
                                             preview_image);
static gboolean gimp_preview_image_update_preview_idle_fun (gpointer data);
static void coordinate_adjustment_value_changed_handler (GtkAdjustment *
                                                         adjustment,
                                                         gpointer data);
static void scale_adjustment_value_changed_handler (GtkAdjustment *
                                                    adjustment,
                                                    gpointer data);
static gboolean gimp_preview_image_expose_handler (GtkWidget * widget,
                                                   GdkEventExpose * event,
                                                   gpointer user_data);
static void gimp_preview_image_allocate_handler (GtkWidget * widget,
                                                 GtkAllocation * allocation,
                                                 gpointer user_data);

/* Constants for transparency checkerboard */
#define CHECK_DARK  ((guchar) (GIMP_CHECK_DARK * 255))
#define CHECK_LIGHT ((guchar) (GIMP_CHECK_LIGHT * 255))

enum
{
  UPDATE_PREVIEW,
  LAST_SIGNAL
};

static guint gimp_preview_image_signals[LAST_SIGNAL] = { 0 };

/* Convert coordinate in preview space to image coordinates */
#define p2i(preview_coordinate, scale) ((gint)((preview_coordinate) / (scale)))

/* Return maximum preview coordinate p such that p2i(p, scale) <=
 * image_coordinate < p2i(p + 1,scale)
 */
static gint
i2p (gint image_coordinate, gdouble scale)
{
  gint p = (gint) ((image_coordinate + 1) * scale) + 1;
  g_assert (p2i (p, scale) > image_coordinate);
  while (p2i (p, scale) > image_coordinate)
    {
      p--;
    };
  return (p);
}


static gint
gpi_x (GimpPreviewImage * preview_image)
{
  return (preview_image->x_adjustment->value);
}

static gint
gpi_y (GimpPreviewImage * preview_image)
{
  return (preview_image->y_adjustment->value);
}

static gdouble
gpi_scale (GimpPreviewImage * preview_image)
{
  return (preview_image->scale_adjustment->value);
}


GType
gimp_preview_image_get_type (void)
{
  static GType preview_type = 0;

  if (!preview_type)
    {
      GTypeInfo preview_info = {
        sizeof (GimpPreviewImageClass),
        (GBaseInitFunc) NULL,
        (GBaseFinalizeFunc) NULL,
        (GClassInitFunc) gimp_preview_image_class_init,
        (GClassFinalizeFunc) NULL,
        (gconstpointer) NULL,   /* class_data */
        sizeof (GimpPreviewImage),
        0,                      /* n_preallocs */
        (GInstanceInitFunc) gimp_preview_image_init,
        (GTypeValueTable *) NULL        /* value_table */
      };

      preview_type =
        g_type_register_static (GTK_TYPE_IMAGE, "GimpPreviewImage",
                                &preview_info, 0);
    }

  return preview_type;
}


/*
 * Initialization, which is done when the preview class is created
 * by GTK's internal mechanisms.
 */
static void
gimp_preview_image_class_init (GimpPreviewImageClass * klass)
{
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);

  gimp_preview_image_signals[UPDATE_PREVIEW] =
    g_signal_new ("update-preview",
                  G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_FIRST,
                  G_STRUCT_OFFSET (GimpPreviewImageClass, update_preview),
                  NULL,
                  NULL,
                  g_cclosure_marshal_VOID__POINTER,
                  G_TYPE_NONE, 1, G_TYPE_POINTER);


  klass->update_preview = NULL;
  object_class->destroy = gimp_preview_image_destroy;
  widget_class->size_request = gimp_preview_image_size_request;
}

/*
 * Initialization, which is done when the preview widget is first created
 * by GTK's internal mechanisms.
 */
static void
gimp_preview_image_init (GimpPreviewImage * preview_image)
{
  g_assert (GIMP_IS_PREVIEW_IMAGE (preview_image));

  preview_image->show_original = TRUE;
  preview_image->previous_scale = 1.0;
  preview_image->idle_fun_busy = FALSE;
  preview_image->render_state = GPI_RENDER_OK;
  preview_image->width = 0;
  preview_image->height = 0;
}

static void
gimp_preview_image_destroy (GtkObject * object)
{
  GimpPreviewImage *preview_image = GIMP_PREVIEW_IMAGE (object);
  GtkObjectClass *parent_class =
    GTK_OBJECT_CLASS (g_type_class_peek_parent
                      (G_OBJECT (preview_image)->g_type_instance.g_class));

  gtk_object_destroy (GTK_OBJECT (preview_image->scale_adjustment));
  gtk_object_destroy (GTK_OBJECT (preview_image->x_adjustment));
  gtk_object_destroy (GTK_OBJECT (preview_image->y_adjustment));

  parent_class->destroy (object);
}


/**
 * gimp_preview_image_new:
 * @drawable: The original drawable that will be shown when the user
 * is scrolling.
 * 
 * Create a new #GimpPreviewImage widget.
 * This call creates a widget with a default size,
 * initial scale = 1.
 * 
 * Return value: The new GimpPreviewImage widget. 
 **/
GtkWidget *
gimp_preview_image_new (GimpDrawable * drawable)
{
  /* Simplified front end that uses all the defaults. */
  return (gimp_preview_image_new_with_args (drawable, -1, -1, 1.0));
}

/**
 * gimp_preview_image_new_with_args:
 * @drawable: The original drawable that will be shown when the user
 * is scrolling.
 * @preview_width: Width of the preview image in pixels. When it is
 * < 0 a default width is used. When it is equal to 0 the preview_image
 * will dynamically resize to fill the allocated area.
 * @preview_height: Heigth of the preview image in pixels. When it is
 * < 0 a default heigth is used. When it is equal to 0 the preview_image
 * will dynamically resize to fill the allocated area.
 * @initial_scale: Initial scale of the preview.
 * 
 * Create a new #GimpPreviewImage widget.
 * 
 * Return value: The new GimpPreviewImage widget. 
 **/
GtkWidget *
gimp_preview_image_new_with_args (GimpDrawable * drawable,
                                  gint preview_width,
                                  gint preview_height, gdouble initial_scale)
{
  GimpPreviewImage *preview_image;

  /* Now allocate the actual preview window. */
  preview_image =
    GIMP_PREVIEW_IMAGE (g_object_new (gimp_preview_image_get_type (), NULL));

  /* Save the drawable info. */
  preview_image->drawable = drawable;
  preview_image->drawable_has_alpha =
    gimp_drawable_has_alpha (drawable->drawable_id);

  /* Set the size */
  if (preview_width < 0)
    {
      preview_width = PREVIEW_DEFAULT_SIZE;
    };
  if (preview_height < 0)
    {
      preview_height = PREVIEW_DEFAULT_SIZE;
    };

  preview_image->requested_width = preview_width;
  preview_image->requested_height = preview_height;

  preview_image->x_adjustment =
    GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 0, 0, 0, 0));
  g_object_ref (preview_image->x_adjustment);
  gtk_object_sink (GTK_OBJECT (preview_image->x_adjustment));
  preview_image->y_adjustment =
    GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 0, 0, 0, 0));
  g_object_ref (preview_image->y_adjustment);
  gtk_object_sink (GTK_OBJECT (preview_image->y_adjustment));
  preview_image->scale_adjustment =
    GTK_ADJUSTMENT (gtk_adjustment_new (1.0, 1e-5, 101, 0.1, 1, 1));
  g_object_ref (preview_image->scale_adjustment);
  gtk_object_sink (GTK_OBJECT (preview_image->scale_adjustment));

  g_signal_connect (G_OBJECT (preview_image->x_adjustment), "value-changed",
                    G_CALLBACK (coordinate_adjustment_value_changed_handler),
                    (gpointer) preview_image);
  g_signal_connect (G_OBJECT (preview_image->y_adjustment), "value-changed",
                    G_CALLBACK (coordinate_adjustment_value_changed_handler),
                    (gpointer) preview_image);
  g_signal_connect (G_OBJECT (preview_image->scale_adjustment),
                    "value-changed",
                    G_CALLBACK (scale_adjustment_value_changed_handler),
                    (gpointer) preview_image);
  g_signal_connect (G_OBJECT (preview_image), "expose_event",
                    G_CALLBACK (gimp_preview_image_expose_handler),
                    (gpointer) NULL);
  g_signal_connect_after (G_OBJECT (preview_image), "size_allocate",
                          G_CALLBACK (gimp_preview_image_allocate_handler),
                          (gpointer) NULL);

  /* Add the idle function that generates the "update-preview"
     signal when needed.
   */
  g_idle_add_full (G_PRIORITY_LOW,
                   gimp_preview_image_update_preview_idle_fun, preview_image,
                   NULL);

  /*
   * Build our initial preview of the drawable data.
   */

  /* Set the scale amount. */
  if (initial_scale <= 0)
    {
      initial_scale = 1.0;
    };
  gimp_preview_image_set_scale (preview_image, initial_scale);
  gimp_preview_image_recompute_sizes (preview_image);

  gimp_preview_image_update (preview_image);

  return GTK_WIDGET (preview_image);
}

/**
 * gimp_preview_image_update:
 * @preview_image: A #GimpPreviewImage.
 * 
 * Plug-ins call this to do an update of the preview area.
 * The preview will generate an "#update-preview" signal.
 * The plug-in's signal handler should then draw in on the preview
 * with one of
 * the drawing functions: gimp_preview_image_draw_unscaled_row(),
 * gimp_preview_image_draw_unscaled_drawable(), gimp_preview_image_draw_scaled_row()
 * or gimp_preview_image_draw_scaled_drawable().
 **/
void
gimp_preview_image_update (GimpPreviewImage * preview_image)
{
  preview_image->render_state = GPI_RENDER_UPDATE_NEEDED;
#ifdef PREVIEW_DEBUG
  printf ("gimp_preview_image_update: update needed\n");
#endif
  gimp_preview_image_force_redraw (preview_image);
}

static gboolean
gimp_preview_image_expose_handler (GtkWidget * widget,
                                   GdkEventExpose * expose,
                                   gpointer user_data)
{
  GimpPreviewImage *preview_image = GIMP_PREVIEW_IMAGE (widget);

  if (preview_image->show_original &&
      (preview_image->render_state == GPI_RENDER_UPDATE_NEEDED))
    {
#ifdef PREVIEW_DEBUG
      printf ("gimp_preview_image_expose_handler: redrawing original\n");
#endif
      gimp_preview_image_redraw_original_image (preview_image);
    }
  return (FALSE);
}


/* Answer a GTK size request */
static void
gimp_preview_image_size_request (GtkWidget * widget,
                                 GtkRequisition * requisition)
{
  GimpPreviewImage *preview_image = GIMP_PREVIEW_IMAGE (widget);
  GtkWidgetClass *parent_class = (GtkWidgetClass *)
    g_type_class_peek_parent (G_OBJECT (preview_image)->g_type_instance.
                              g_class);

#ifdef PREVIEW_DEBUG
  printf ("Received size request request for %d x %d\n", requisition->width,
          requisition->height);
#endif
  if (preview_image->requested_width > 0)
    {
      requisition->width =
        MAX (1,
             preview_image->requested_width +
             GTK_MISC (preview_image)->xpad * 2);
    }
  if (preview_image->requested_height > 0)
    {
      requisition->height =
        MAX (1,
             preview_image->requested_height +
             GTK_MISC (preview_image)->ypad * 2);
    }
  /* Don't really know if this is necessary */
  widget->requisition = *requisition;

  /* Now call the parent's size_request */
  parent_class->size_request (widget, requisition);

#ifdef PREVIEW_DEBUG
  printf ("Return size request request %d x %d\n", requisition->width,
          requisition->height);
#endif
}

static void
gimp_preview_image_allocate_handler (GtkWidget * widget,
                                     GtkAllocation * allocation,
                                     gpointer user_data)
{
  GimpPreviewImage *preview_image = GIMP_PREVIEW_IMAGE (widget);

  widget->allocation = *allocation;
  gimp_preview_image_recompute_sizes (preview_image);
}

static void
gimp_preview_image_force_redraw (GimpPreviewImage * preview_image)
{
  gtk_widget_queue_draw (GTK_WIDGET (preview_image));
}


void
gimp_preview_image_generate_update_event (GimpPreviewImage * preview_image)
     /* Signal the user that the preview must be updated */
{
  const gdouble scale = gpi_scale (preview_image);
  const gint image_x = p2i (gpi_x (preview_image), scale);
  const gint image_y = p2i (gpi_y (preview_image), scale);

  const gint image_width =
    CLAMP (p2i (gpi_x (preview_image) + preview_image->width - 1, scale) + 1 -
           image_x,
           0,
           preview_image->drawable->width - image_x);

  const gint image_height =
    CLAMP (p2i (gpi_y (preview_image) + preview_image->height - 1, scale) +
           1 - image_y,
           0,
           preview_image->drawable->height - image_y);

  GimpPreviewImageUpdateEvent preview_event;

  preview_event.scale = gpi_scale (preview_image);
  preview_event.image_x = image_x;
  preview_event.image_y = image_y;
  preview_event.image_width = image_width;
  preview_event.image_height = image_height;
  preview_event.preview_x = gpi_x (preview_image);
  preview_event.preview_y = gpi_y (preview_image);
  preview_event.preview_width = preview_image->width;
  preview_event.preview_height = preview_image->height;
  g_signal_emit (GTK_OBJECT (preview_image),
                 gimp_preview_image_signals[UPDATE_PREVIEW], 0,
                 &preview_event);
  gimp_preview_image_force_redraw (preview_image);
}

/* This function can be called from the glib idle loop */
gboolean
gimp_preview_image_update_preview_idle_fun (gpointer data)
{
  GimpPreviewImage *preview_image = GIMP_PREVIEW_IMAGE (data);

  if (!preview_image->idle_fun_busy)
    {
      preview_image->idle_fun_busy = TRUE;
      while (preview_image->render_state == GPI_RENDER_UPDATE_NEEDED)
        {
#ifdef PREVIEW_DEBUG
          printf ("gimp_preview_image_preview_idle_fun: rendering\n");
#endif
          preview_image->render_state = GPI_RENDER_RENDERING;
          gimp_preview_image_generate_update_event (preview_image);
        }
      preview_image->render_state = GPI_RENDER_OK;
      preview_image->idle_fun_busy = FALSE;
    }
  return (TRUE);
}

static void
coordinate_adjustment_value_changed_handler (GtkAdjustment * adjustment,
                                             gpointer data)
{
  GimpPreviewImage *preview_image = GIMP_PREVIEW_IMAGE (data);

  g_assert (GTK_IS_ADJUSTMENT (adjustment));

  gimp_preview_image_update (preview_image);
}

/* Recompute the size and location of the display after a scale change
 * or new allocation.
 * Try to keep the same location in the image at the centre of the preview.
 * This function is also used for initializing the preview.
 */
static void
gimp_preview_image_recompute_sizes (GimpPreviewImage * preview_image)
{

  /* The center of the preview in image coordinates.
   * We try to keep the center at the same position when changing scales
   */
  gdouble old_scale = preview_image->previous_scale;
  gdouble new_scale = gpi_scale (preview_image);

  const gint im_center_x =
    p2i (gpi_x (preview_image) + preview_image->width / 2.0, old_scale);
  const gint im_center_y =
    p2i (gpi_y (preview_image) + preview_image->height / 2.0, old_scale);

  const gint image_width =
    i2p (preview_image->drawable->width - 1, new_scale) + 1;
  const gint image_height =
    i2p (preview_image->drawable->height - 1, new_scale) + 1;

  gimp_preview_image_set_size (preview_image, image_width, image_height);
  gimp_preview_image_set_position (preview_image,
                                   CLAMP (i2p (im_center_x, new_scale) -
                                          preview_image->width / 2,
                                          0,
                                          MAX (0,
                                               image_width -
                                               preview_image->width)),
                                   CLAMP (i2p (im_center_y, new_scale) -
                                          preview_image->height / 2,
                                          0,
                                          MAX (0,
                                               image_height -
                                               preview_image->height)));

  /* Assert that bottom-right point of preview maps to a point in the drawable */
  g_assert (p2i
            (gpi_x (preview_image) + preview_image->width - 1,
             gpi_scale (preview_image)) < preview_image->drawable->width);
  g_assert (p2i
            (gpi_y (preview_image) + preview_image->height - 1,
             gpi_scale (preview_image)) < preview_image->drawable->height);
}

static void
scale_adjustment_value_changed_handler (GtkAdjustment * adjustment,
                                        gpointer data)
{
  GimpPreviewImage *preview_image = GIMP_PREVIEW_IMAGE (data);

  g_assert (GTK_IS_ADJUSTMENT (adjustment));
  gimp_preview_image_recompute_sizes (preview_image);
  preview_image->previous_scale = gpi_scale (preview_image);
  gimp_preview_image_update (preview_image);
}


static void
update_adjustment (GtkAdjustment * adjustment, gint max_coordinate,
                   gint preview_size)
{
  const gdouble lower = 0.0;
  const gdouble upper = max_coordinate - 1;
  const gdouble step_increment = 1.0;
  const gdouble page_increment = preview_size;
  const gdouble page_size = preview_size;
  const gdouble value = CLAMP (adjustment->value, lower, max_coordinate);
  const gboolean value_changed = (value != adjustment->value);
  const gboolean changed =
    ((lower != adjustment->lower) ||
     (upper != adjustment->upper) ||
     (step_increment != adjustment->step_increment) ||
     (page_increment != adjustment->page_increment) ||
     (page_size != adjustment->page_size) ||
     (lower != adjustment->lower) ||
     (lower != adjustment->lower) || (lower != adjustment->lower));

  adjustment->lower = lower;
  adjustment->upper = upper;
  adjustment->step_increment = step_increment;
  adjustment->page_increment = page_increment;
  adjustment->page_size = page_size;
  adjustment->value = value;

  if (changed)
    {
      gtk_adjustment_changed (adjustment);
    }
  if (value_changed)
    {
      gtk_adjustment_value_changed (adjustment);
    }
}

/* Compute new size for the preview image and allocate buffers */
void
gimp_preview_image_set_size (GimpPreviewImage * preview_image,
                             gint image_width, gint image_height)
{
  gint real_width, real_height;

  real_width =
    MIN (GTK_WIDGET (preview_image)->allocation.width
         - GTK_MISC (preview_image)->xpad * 2, image_width);
  if (preview_image->requested_width > 0)
    {
      real_width = MIN (preview_image->requested_width, image_width);
    };


  real_height =
    MIN (GTK_WIDGET (preview_image)->allocation.height
         - GTK_MISC (preview_image)->ypad * 2, image_height);
  if (preview_image->requested_height > 0)
    {
      real_height = MIN (preview_image->requested_height, image_height);
    };


  if ((preview_image->width != real_width)
      || (preview_image->height != real_height))
    {
      GdkPixbuf *pixbuf;

      /* This must be done here before changing the pixbuf
         otherwise we get an infinite recursion */
      preview_image->width = real_width;
      preview_image->height = real_height;

      pixbuf =
        gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, real_width,
                        real_height);
      gtk_image_set_from_pixbuf (GTK_IMAGE (preview_image), pixbuf);
      g_object_unref (pixbuf);
      preview_image->width = real_width;
      preview_image->height = real_height;

      gimp_preview_image_update (preview_image);
    }

  update_adjustment (preview_image->x_adjustment, image_width, real_width);
  update_adjustment (preview_image->y_adjustment, image_height, real_height);
}



/* Low level function to draw a single row on the pixbuf */
static void
gimp_preview_image_pixbuf_draw_row (GdkPixbuf * pixbuf, guchar * data,
                                    gint x, gint y, gint w)
{
  const gint n_chars_per_pixel =
    gdk_pixbuf_get_n_channels (pixbuf) *
    (gdk_pixbuf_get_bits_per_sample (pixbuf) / 8);
  guchar *dest;

  g_assert ((x >= 0) && (x < gdk_pixbuf_get_width (pixbuf)));
  g_assert ((w >= 0) && ((x + w) <= gdk_pixbuf_get_width (pixbuf)));
  g_assert ((y >= 0) && (y < gdk_pixbuf_get_height (pixbuf)));

  dest =
    gdk_pixbuf_get_pixels (pixbuf) + y * gdk_pixbuf_get_rowstride (pixbuf) +
    x * n_chars_per_pixel;
  memcpy (dest, data, w * n_chars_per_pixel);
}

/* Low level function to draw a single row on the preview */
static void
gimp_preview_image_draw_row (GimpPreviewImage * preview_image, guchar * data,
                             gint x, gint y, gint w)
{
  GdkPixbuf *pixbuf;

  pixbuf = gtk_image_get_pixbuf (GTK_IMAGE (preview_image));
  gimp_preview_image_pixbuf_draw_row (pixbuf, data, x, y, w);
}

/**
 * gimp_preview_image_draw_scaled_row:
 * @preview_image: the #GimpPreviewImage
 * @type: the format of the data (e.g. %GIMP_RGBA_IMAGE).
 * @row:the relative number of the row within the preview.
 * The top row of the preview is number 0.
 * @width: the number of pixels in this row.
 * @data: pixels for the preview. It must have a length of width pixels
 *
 * Draw one scaled row of data in the preview.
 * This function handles conversion to RGB and checkerboarding of
 * transparent areas.
 *
 * It is only possible to draw from within an "update-preview"
 * signal handler when the user has not scrolled or zoomed
 * after the invocation of this signal handler.
 * In all other cases drawing on the preview is not
 * allowed and all drawing requests are ignored.
 * Rendering functions should halt immediately when
 * this function returns %FALSE, because the results of
 * this rendering function will be ignored by the preview.
 *
 * Return value: %TRUE when the drawing on the preview is allowed.
 * When this function returns %FALSE the rendering should
 * halt immediately.
 **/
gboolean
gimp_preview_image_draw_scaled_row (GimpPreviewImage * preview_image,
                                    GimpImageType type,
                                    gint row, gint width, const guchar * data)
{
  /* This implementation uses explicit subscripting with loop variables.
   * This makes this version faster than a version that uses pointers
   * to step through source and destination!
   */
  /* When the event-id is not current return FALSE to indicate
   * the plug-in that it can stop with updates for this event.
   */
  const gboolean return_status =
    preview_image->render_state == GPI_RENDER_RENDERING;

  if (return_status)
    {
      const gint width = preview_image->width;
      guchar *buffer = g_new (guchar, 3 * width);
      const gint check_offset =
        ((gpi_y (preview_image) + row) & GIMP_CHECK_SIZE) ?
        gpi_x (preview_image) : gpi_x (preview_image) + GIMP_CHECK_SIZE;
      gint j;

      if ((row < 0) || (row >= preview_image->height))
        {
          g_free (buffer);
          return (return_status);
        }

      switch (type)
        {
        case GIMP_RGBA_IMAGE:
          {
            for (j = width - 1; j >= 0; j--)
              {
                const guchar check = ((j + check_offset) & GIMP_CHECK_SIZE) ?
                  CHECK_LIGHT : CHECK_DARK;

                buffer[3 * j] =
                  check +
                  (((data[4 * j + 0] - check) * data[4 * j + 3]) >> 8);
                buffer[3 * j + 1] =
                  check +
                  (((data[4 * j + 1] - check) * data[4 * j + 3]) >> 8);
                buffer[3 * j + 2] =
                  check +
                  (((data[4 * j + 2] - check) * data[4 * j + 3]) >> 8);
              }
          }
          break;
        case GIMP_RGB_IMAGE:
          {
            /* Yes, you're right. We could use memcpy here */
            for (j = width - 1; j >= 0; j--)
              {
                buffer[3 * j] = data[3 * j];
                buffer[3 * j + 1] = data[3 * j + 1];
                buffer[3 * j + 2] = data[3 * j + 2];
              }
          }
          break;
        case GIMP_GRAYA_IMAGE:
          {
            for (j = width - 1; j >= 0; j--)
              {
                const guchar check = ((j + check_offset) & GIMP_CHECK_SIZE) ?
                  CHECK_LIGHT : CHECK_DARK;

                buffer[3 * j] =
                  buffer[3 * j + 1] =
                  buffer[3 * j + 2] =
                  check + (((data[2 * j] - check) * data[2 * j + 1]) >> 8);
              }
          }
          break;
        case GIMP_GRAY_IMAGE:
          {
            for (j = width - 1; j >= 0; j--)
              {

                buffer[3 * j] =
                  buffer[3 * j + 1] = buffer[3 * j + 2] = data[j];
              }
          }
          break;
        default:
          /* Type argument is wrong */
          g_assert_not_reached ();
          break;
        }

      gimp_preview_image_draw_row (preview_image, buffer, 0, row, width);
      g_free (buffer);
    }
  return (return_status);
}

/**
 * gimp_preview_image_draw_unscaled_row:
 * @preview_image: the #GimpPreviewImage
 * @type: the format of the data (e.g. %GIMP_RGBA_IMAGE).
 * @row:row is the relative position of the row w.r.t. preview_event->image_y.
 * The top row has number 0.
 * @width: the number of pixels in this row.
 * @data: pixels for the preview. The first pixel in @data has x-coordinate
 * preview_event->image_x in the image. @data must have a length
 * of @width pixels.
 *
 * Draw one unscaled row of data in the preview.
 * This function handles scaling, conversion to RGB and checkerboarding of
 * transparent areas.
 * A nice feature is that this function will draw several lines of
 * the preview when scale > 1.
 *
 *
 * It is only possible to draw from within an "update-preview"
 * signal handler when the user has not scrolled or zoomed
 * after the invocation of this signal handler.
 * In all other cases drawing on the preview is not
 * allowed and all drawing requests are ignored.
 * Rendering functions should halt immediately when
 * this function returns %FALSE, because the results of
 * this rendering function will be ignored by the preview.
 *
 * Return value: %TRUE when the drawing on the preview is allowed.
 * When this function returns %FALSE the rendering should
 * halt immediately.
 **/
gboolean
gimp_preview_image_draw_unscaled_row (GimpPreviewImage * preview_image,
                                      GimpImageType type,
                                      gint row, gint width,
                                      const guchar * const data)
{
  /*
     There is a small complication with transparent data when scale > 1
     that is caused by the checkerboarding. Since there are two different
     checkerboard patterns we may have to compute two different versions
     for the preview.
     This implementation uses explicit subscripting with loop variables.
     This makes this version faster than a version that uses pointers
     to step through source and destination!
   */

  const gboolean return_status =
    preview_image->render_state == GPI_RENDER_RENDERING;

  if (return_status)
    {
      const gint preview_width = preview_image->width;
      /* Indicate whether we must compute the odd and/or the even version 
         for checkerboarding transparent areas */
      gboolean need_check_idx[2] = { FALSE, FALSE };
      /* Buffers for the odd and even version of the checkboarded line.
       * Image-types without alpha only use the first buffer.
       */
      guchar *buffer[2] = { NULL, NULL };
      gint check_idx;
      const gdouble scale = gpi_scale (preview_image);
      const gint preview_x = gpi_x (preview_image);
      const gint preview_y = gpi_y (preview_image);
      const gint image_x = p2i (preview_x, scale);
      const gint image_y = p2i (preview_y, scale);

      /* Compute the range of the lines in the preview that we can draw */
      const gint y_start = CLAMP (i2p (image_y + row - 1, scale) + 1,
                                  preview_y,
                                  preview_y + preview_image->height - 1);
      /* This is the end of the range. This line will not be drawn. */
      const gint y_end = CLAMP (i2p (image_y + row, scale) + 1,
                                preview_y,
                                preview_y + preview_image->height);

      gint x, y;

      g_return_val_if_fail (p2i (preview_x + preview_width - 1, scale)
                            <= (image_x + width - 1), FALSE);

      if (y_start >= y_end)
        {
          /* This row is not used in the preview */
          return (return_status);
        }

      if ((type == GIMP_RGBA_IMAGE) || (type == GIMP_GRAYA_IMAGE))
        {
          /* Determine which versions of the checkerboard are needed */
          for (y = y_start; y < y_end; y++)
            {
              need_check_idx[(y & GIMP_CHECK_SIZE) ? 1 : 0] = TRUE;
            }
          for (check_idx = 0; check_idx < 2; check_idx++)
            if (need_check_idx[check_idx])
              buffer[check_idx] = g_new (guchar, 3 * preview_width);
        }
      else
        {
          need_check_idx[0] = TRUE;
          buffer[0] = g_new (guchar, 3 * preview_width);
        }

      /* Scale and checkerboard the image data */
      switch (type)
        {
        case GIMP_RGBA_IMAGE:
          for (check_idx = 0; check_idx < 2; check_idx++)
            {
              if (need_check_idx[check_idx])
                {
                  guchar *const dest_ptr = buffer[check_idx];
                  const gint check_offset =
                    preview_x + GIMP_CHECK_SIZE * (1 - check_idx);

                  for (x = preview_width - 1; x >= 0; x--)
                    {
                      const guchar check =
                        ((x +
                          check_offset) & GIMP_CHECK_SIZE) ? CHECK_LIGHT :
                        CHECK_DARK;
                      const guchar *const src_ptr =
                        data + 4 * (p2i (preview_x + x, scale) - image_x);

                      dest_ptr[3 * x + 0] =
                        check + (((src_ptr[0] - check) * src_ptr[3]) >> 8);
                      dest_ptr[3 * x + 1] =
                        check + (((src_ptr[1] - check) * src_ptr[3]) >> 8);
                      dest_ptr[3 * x + 2] =
                        check + (((src_ptr[2] - check) * src_ptr[3]) >> 8);
                    }
                }
            }
          break;
        case GIMP_RGB_IMAGE:
          {
            guchar *const dest_ptr = buffer[0];

            for (x = preview_width - 1; x >= 0; x--)
              {
                const guchar *const src_ptr = data +
                  3 * (p2i (preview_x + x, scale) - image_x);

                dest_ptr[3 * x + 0] = src_ptr[0];
                dest_ptr[3 * x + 1] = src_ptr[1];
                dest_ptr[3 * x + 2] = src_ptr[2];
              }
          }
          break;
        case GIMP_GRAYA_IMAGE:
          for (check_idx = 0; check_idx < 2; check_idx++)
            {
              if (need_check_idx[check_idx])
                {
                  guchar *const dest_ptr = buffer[check_idx];
                  const gint check_offset =
                    preview_x + GIMP_CHECK_SIZE * (1 - check_idx);

                  for (x = preview_width - 1; x >= 0; x--)
                    {
                      const guchar check =
                        ((x +
                          check_offset) & GIMP_CHECK_SIZE) ? CHECK_LIGHT :
                        CHECK_DARK;
                      const guchar *const src_ptr =
                        data + 2 * (p2i (preview_x + x, scale) - image_x);

                      dest_ptr[3 * x + 0] =
                        dest_ptr[3 * x + 1] =
                        dest_ptr[3 * x + 2] =
                        check + (((src_ptr[0] - check) * src_ptr[1]) >> 8);
                    }
                }
            }
          break;
        case GIMP_GRAY_IMAGE:
          {
            guchar *const dest_ptr = buffer[0];

            for (x = preview_width - 1; x >= 0; x--)
              {
                const guchar *const src_ptr = data +
                  p2i (preview_x + x, scale) - image_x;

                dest_ptr[3 * x + 0] =
                  dest_ptr[3 * x + 1] = dest_ptr[3 * x + 2] = src_ptr[0];
              }
          }
          break;
        default:
          /* Type argument is wrong */
          g_assert_not_reached ();
          break;
        }

      /* Now draw the scaled and checkerboarded rows into the pixmap */
      for (y = y_start; y < y_end; y++)
        {
          /* For images without alpha always use buffer[0],
           * else use the correct checkerboarded version
           * */
          const gint buffer_idx =
            ((type == GIMP_RGB_IMAGE) || (type == GIMP_GRAY_IMAGE))
            ? 0 : (y & GIMP_CHECK_SIZE) ? 1 : 0;

          g_assert (need_check_idx[buffer_idx]);
          g_assert (buffer[buffer_idx]);

          gimp_preview_image_draw_row (preview_image,
                                       buffer[buffer_idx],
                                       0, y - preview_y, preview_width);
        }

      /* g_free handles NULL pointers gracefully */
      g_free (buffer[0]);
      g_free (buffer[1]);
    }
  return (return_status);
}


/**
 * gimp_preview_image_draw_scaled_drawable:
 * @preview_image: a #GimpPreviewImage.
 * @drawable: A #GimpDrawable that contains the scaled rendered image.
 * 
 * Draw a scaled drawable in the preview. 
 * 
 *
 * It is only possible to draw from within an "update-preview"
 * signal handler when the user has not scrolled or zoomed
 * after the invocation of this signal handler.
 * In all other cases drawing on the preview is not
 * allowed and all drawing requests are ignored.
 * Rendering functions should halt immediately when
 * this function returns %FALSE, because the results of
 * this rendering function will be ignored by the preview.
 *
 * Return value: %TRUE when the drawing on the preview is allowed.
 * When this function returns %FALSE the rendering should
 * halt immediately.
 **/
gboolean
gimp_preview_image_draw_scaled_drawable (GimpPreviewImage * preview_image,
                                         GimpDrawable * drawable)
{
  const gboolean return_status =
    preview_image->render_state == GPI_RENDER_RENDERING;

  if (return_status)
    {
      const gint width = preview_image->width;
      const gint height = preview_image->height;

      GimpPixelRgn pixel_region;
      gint y;
      guchar *row_buf =
        g_malloc (width * gimp_drawable_bytes (drawable->drawable_id));

      g_assert (width <= drawable->width);
      g_assert (height <= drawable->height);
      gimp_pixel_rgn_init (&pixel_region, drawable, 0, 0, width,
                           height, FALSE, FALSE);
      for (y = 0; y < height; y++)
        {
          gimp_pixel_rgn_get_row (&pixel_region, row_buf, 0, y, width);
          gimp_preview_image_draw_scaled_row (preview_image,
                                              gimp_drawable_type (drawable->
                                                                  drawable_id),
                                              y, width, row_buf);
        }
      g_free (row_buf);
    }
  return (return_status);
}

/**
 * gimp_preview_image_draw_unscaled_drawable_with_offset:
 * @preview_image: a #GimpPreviewImage.
 * @drawable: A #GimpDrawable that contains the unscaled rendered image.
 * @offset_x: Coordinate of the left of the preview in the @drawable.
 * @offset_y: Coordinate of the top of the preview in the @drawable.

 * Draw an unscaled drawable in the preview. 
 * 
 * It is only possible to draw from within an "update-preview"
 * signal handler when the user has not scrolled or zoomed
 * after the invocation of this signal handler.
 * In all other cases drawing on the preview is not
 * allowed and all drawing requests are ignored.
 * Rendering functions should halt immediately when
 * this function returns %FALSE, because the results of
 * this rendering function will be ignored by the preview.
 *
 * Return value: %TRUE when the drawing on the preview is allowed.
 * When this function returns %FALSE the rendering should
 * halt immediately.
 **/
static void
draw_unscaled_drawable_with_offset (GimpPreviewImage * preview_image,
                                    GimpDrawable * drawable,
                                    gint offset_x, gint offset_y)
{
  GimpPixelRgn region;
  guchar *image_data = NULL;
  guchar *row_buffer = NULL;

  const gint preview_x = gpi_x (preview_image);
  const gint preview_y = gpi_y (preview_image);
  const gdouble scale = gpi_scale (preview_image);
  /* Coordinates within the image of top left pixel in preview */
  const gint image_x = p2i (preview_x, scale);
  const gint image_y = p2i (preview_y, scale);


  /* Determine the available area in the drawable */
  const gint image_width =
    CLAMP (p2i (preview_x + preview_image->width - 1, scale) + 1 - offset_x,
           0,
           drawable->width - offset_x);

  const gint image_height =
    CLAMP (p2i (preview_y + preview_image->height - 1, scale) + 1 - offset_y,
           0,
           drawable->height - offset_y);

  /* This is the area in the preview that we can actually draw */
  const gint width = CLAMP (preview_image->width,
                            0,
                            i2p (image_x + image_width - 1,
                                 scale) + 1 - preview_x);
  const gint height = CLAMP (preview_image->height,
                             0,
                             i2p (image_y + image_height - 1,
                                  scale) + 1 - preview_y);

  g_assert ((offset_x >= 0) && (offset_x < drawable->width));
  g_assert ((offset_y >= 0) && (offset_y < drawable->height));
  g_assert (offset_x + image_width <= drawable->width);
  g_assert (offset_y + image_height <= drawable->height);
  gimp_pixel_rgn_init (&region, drawable,
                       offset_x, offset_y, image_width, image_height, FALSE,
                       FALSE);

  /* This should happen during initialization to speed things up on large images. */
  row_buffer = g_new0 (guchar, 3 * preview_image->width);
  image_data = g_new0 (guchar, drawable->bpp * image_width * image_height);

  gimp_pixel_rgn_get_rect (&region, image_data,
                           offset_x, offset_y, image_width, image_height);


  /* Scale and convert to RGB */
  if (drawable->bpp == 4)
    {
      gint dy;
      const gint bpp = 4;

      /* For each row of the preview ... */
      for (dy = 0; dy < height; dy++)
        {
          /* Calculate: starting position in drawable */
          gint dx;
          /* y coordinate in image */
          const gint sy = p2i (dy + preview_y, scale);
          const guchar *const src_ptr =
            image_data + ((sy - image_y) * image_width - image_x) * bpp;
          const gint check_offset =
            ((preview_y + dy) & GIMP_CHECK_SIZE) ? preview_x : preview_x +
            GIMP_CHECK_SIZE;

          for (dx = 0; dx < width; dx++)
            {
              /* x coordinate in image */
              const gint sx = p2i (dx + preview_x, scale);
              const guchar check = ((dx + check_offset) & GIMP_CHECK_SIZE) ?
                CHECK_LIGHT : CHECK_DARK;

              row_buffer[dx * 3 + 0] =
                check +
                (((src_ptr[sx * bpp + 0] - check) * src_ptr[sx * bpp +
                                                            3]) >> 8);
              row_buffer[dx * 3 + 1] =
                check +
                (((src_ptr[sx * bpp + 1] - check) * src_ptr[sx * bpp +
                                                            3]) >> 8);
              row_buffer[dx * 3 + 2] =
                check +
                (((src_ptr[sx * bpp + 2] - check) * src_ptr[sx * bpp +
                                                            3]) >> 8);
            }
          gimp_preview_image_draw_row (preview_image, row_buffer, 0, dy,
                                       width);
        }
    }
  else if (drawable->bpp == 3)
    {
      gint dy;
      const gint bpp = 3;

      /* For each row of the preview ... */
      for (dy = 0; dy < height; dy++)
        {
          /* Calculate: starting position in drawable */
          gint dx;
          /* y coordinate in image */
          const gint sy = p2i (dy + preview_y, scale);

          const guchar *const src_ptr =
            image_data + ((sy - image_y) * image_width - image_x) * bpp;

          for (dx = 0; dx < width; dx++)
            {
              /* x coordinate in image */
              const gint sx = p2i (dx + preview_x, scale);

              row_buffer[dx * 3 + 0] = src_ptr[sx * bpp + 0];
              row_buffer[dx * 3 + 1] = src_ptr[sx * bpp + 1];
              row_buffer[dx * 3 + 2] = src_ptr[sx * bpp + 2];
            }
          gimp_preview_image_draw_row (preview_image, row_buffer, 0, dy,
                                       width);
        }

    }
  else if (drawable->bpp == 2)
    {
      gint dy;
      const gint bpp = 2;

      for (dy = 0; dy < height; dy++)
        {
          /* Calculate: starting position in drawable */
          gint dx;
          /* y coordinate in image */
          const gint sy = p2i (dy + preview_y, scale);

          const guchar *const src_ptr =
            image_data + ((sy - image_y) * image_width - image_x) * bpp;
          const gint check_offset =
            ((preview_y + dy) & GIMP_CHECK_SIZE) ? preview_x : preview_x +
            GIMP_CHECK_SIZE;

          for (dx = 0; dx < width; dx++)
            {
              /* x coordinate in image */
              const gint sx = p2i (dx + preview_x, scale);
              const guchar check = ((dx + check_offset) & GIMP_CHECK_SIZE) ?
                CHECK_LIGHT : CHECK_DARK;

              row_buffer[dx * 3 + 0] =
                row_buffer[dx * 3 + 1] =
                row_buffer[dx * 3 + 2] =
                check +
                (((src_ptr[sx * bpp + 0] - check) * src_ptr[sx * bpp +
                                                            1]) >> 8);
            }
          gimp_preview_image_draw_row (preview_image, row_buffer, 0, dy,
                                       width);
        }
    }
  else if (drawable->bpp == 1)
    {
      gint dy;
      const gint bpp = 1;

      /* For each row of the preview ... */
      for (dy = 0; dy < height; dy++)
        {
          /* Calculate: starting position in drawable */
          gint dx;
          /* y coordinate in image */
          const gint sy = p2i (dy + preview_y, scale);

          const guchar *const src_ptr =
            image_data + ((sy - image_y) * image_width - image_x) * bpp;

          for (dx = 0; dx < width; dx++)
            {
              /* x coordinate in image */
              const gint sx = p2i (dx + preview_x, scale);

              row_buffer[dx * 3 + 0] =
                row_buffer[dx * 3 + 1] =
                row_buffer[dx * 3 + 2] = src_ptr[sx * bpp + 0];
            }
          gimp_preview_image_draw_row (preview_image, row_buffer, 0, dy,
                                       width);
        }
    }

  g_free (image_data);
  g_free (row_buffer);
}

/* Redraw the preview with the scaled original image. 
 */
static void
gimp_preview_image_redraw_original_image (GimpPreviewImage * preview_image)
{
  const gdouble scale = gpi_scale (preview_image);
  const gint image_x = p2i (gpi_x (preview_image), scale);
  const gint image_y = p2i (gpi_y (preview_image), scale);

  draw_unscaled_drawable_with_offset (preview_image,
                                      preview_image->drawable,
                                      image_x, image_y);
}

/**
 * gimp_preview_image_draw_unscaled_drawable:
 * @preview_image: a #GimpPreviewImage.
 * @drawable: A #GimpDrawable that contains the unscaled rendered image.
 * 
 * Draw an unscaled drawable in the preview. 
 * 
 * It is only possible to draw from within an "update-preview"
 * signal handler when the user has not scrolled or zoomed
 * after the invocation of this signal handler.
 * In all other cases drawing on the preview is not
 * allowed and all drawing requests are ignored.
 * Rendering functions should halt immediately when
 * this function returns %FALSE, because the results of
 * this rendering function will be ignored by the preview.
 *
 * Return value: %TRUE when the drawing on the preview is allowed.
 * When this function returns %FALSE the rendering should
 * halt immediately.
 **/
gboolean
gimp_preview_image_draw_unscaled_drawable (GimpPreviewImage * preview_image,
                                           GimpDrawable * drawable)
{
  const gboolean return_status =
    preview_image->render_state == GPI_RENDER_RENDERING;

  if (return_status)
    {
      draw_unscaled_drawable_with_offset (preview_image, drawable, 0, 0);
    }
  return (return_status);
}

/**
 * gtk_adjustment_set_correct_value:
 * @adjustment: A #GtkAdjustment.
 * @value: The new value.
 * 
 * This is a replacement for gtk_adjustment_set_value() that
 * uses correct range checks.
 * When a #GtkAdjustment is used with a #GtkRange the legal
 * range of its value is
 * lower .. (upper - page_size).
 * But gtk_adjustment_set_value() assumes that the legal range
 * is lower .. upper.
 * This function fixes that bug.
 * 
 * This function should be removed if the bug is fixed in Gtk.
 **/
void
gtk_adjustment_set_correct_value (GtkAdjustment * adjustment, gdouble value)
{
  g_return_if_fail (GTK_IS_ADJUSTMENT (adjustment));

  const gdouble correct_value = CLAMP (value, adjustment->lower,
                                       adjustment->upper -
                                       adjustment->page_size);

  gtk_adjustment_set_value (adjustment, correct_value);
}



/**
 * gimp_preview_image_set_position:
 * @preview_image: A #GimpPreviewImage.
 * @x_pos: The left side of the preview in #preview-coordinates.
 * @y_pos: The top side of the preview in #preview-coordinates.
 * 
 * Set the position of the area that is show in the preview.
 **/
void
gimp_preview_image_set_position (GimpPreviewImage * preview_image,
                                 gint x_pos, gint y_pos)
{
  gtk_adjustment_set_correct_value (preview_image->x_adjustment, x_pos);
  gtk_adjustment_set_correct_value (preview_image->y_adjustment, y_pos);
}

/**
 * gimp_preview_image_get_preview_position:
 * @preview_image: A #GimpPreviewImage. 
 * @prev_x_pos: Left of the preview in #preview-coordinates.
 * @prev_y_pos: Top of the preview in #preview-coordinates.
 * @prev_width: Width of the preview.
 * @prev_height: Height of the preview.
 * 
 * Return the size and position of the preview in #preview-coordinates.
 **/
void
gimp_preview_image_get_preview_position (GimpPreviewImage * preview_image,
                                         gint * prev_x_pos, gint * prev_y_pos,
                                         gint * prev_width,
                                         gint * prev_height)
{
  *prev_x_pos = gpi_x (preview_image);
  *prev_y_pos = gpi_y (preview_image);
  *prev_width = preview_image->width;
  *prev_height = preview_image->height;
}

/**
 * gimp_preview_image_get_image_position:
 * @preview_image: A #GimpPreviewImage.
 * @im_x_pos: Left of the preview in #image-coordinates.
 * @im_y_pos: Top of the preview in #image-coordinates.
 * @im_width: Width of the preview in #image-coordinates.
 * @im_height: Heigth of the preview in #image-coordinates.
 * 
 * Return the size and position of the preview in #image-coordinates.
 * The preview shows a scaled version of this area of the original or
 * rendered image.
 **/
void
gimp_preview_image_get_image_position (GimpPreviewImage * preview_image,
                                       gint * im_x_pos, gint * im_y_pos,
                                       gint * im_width, gint * im_height)
{
  const gdouble scale = gpi_scale (preview_image);
  *im_x_pos = p2i (gpi_x (preview_image), scale);
  *im_y_pos = p2i (gpi_y (preview_image), scale);

  *im_width =
    CLAMP (p2i (gpi_x (preview_image) + preview_image->width - 1, scale) + 1 -
           *im_x_pos, 0, preview_image->drawable->width - *im_x_pos);

  *im_height =
    CLAMP (p2i (gpi_y (preview_image) + preview_image->height - 1, scale) +
           1 - *im_y_pos, 0, preview_image->drawable->height - *im_y_pos);
}

/**
 * gimp_preview_image_set_scale:
 * @preview_image: A #GimpPreviewImage.
 * @scale: The new scale factor.
 * 
 * Set the current scale factor for the preview.
 **/
void
gimp_preview_image_set_scale (GimpPreviewImage * preview_image, gdouble scale)
{
  gtk_adjustment_set_correct_value (preview_image->scale_adjustment, scale);
}

/**
 * gimp_preview_image_get_scale:
 * @preview_image: A #GimpPreviewImage. 
 * 
 * Get the current scale factor for the preview.
 * 
 * Return value: The scale factor.
 **/
gdouble
gimp_preview_image_get_scale (GimpPreviewImage * preview_image)
{
  return (gpi_scale (preview_image));
}

/**
 * gimp_preview_image_is_drawing_allowed:
 * @preview_image: A #GimpPreviewImage.
 * 
 * Return if it is possible to draw on the preview image.
 * It is only possible to draw from within an "update-preview"
 * signal handler when the user has not scrolled or zoomed
 * after the invocation of this signal handler.
 * In all other cases drawing on the preview is not
 * allowed and all drawing requests are ignored.
 * Rendering functions should halt immediately when
 * this function returns %FALSE, because the results of
 * this rendering function will be ignored by the preview.
 * 
 * Return value: %TRUE when the drawing on the preview is allowed.
 **/
gboolean
gimp_preview_image_is_drawing_allowed (GimpPreviewImage * preview_image)
{
  return (preview_image->render_state == GPI_RENDER_RENDERING);
}

/**
 * gimp_preview_image_get_x_adjustment:
 * @preview_image: A #GimpPreviewImage.
 * 
 * Return the adjustment that manages the x-coordinate of
 * the preview position. The value of this adjustment
 * is expressed in #preview-coordinates.
 *
 * Applications should only change the value of the adjustment,
 * the other properties of the adjustment are managed by the widget.
 * When you try to set the value directly (i.e. not via a GtkScale widget)
 * you should use gtk_adjustment_set_correct_value() (don't use
 * gtk_adjustment_set_value() because it is buggy).
 * 
 * Return value: The #GtkAdjustment.
 **/
GtkAdjustment *
gimp_preview_image_get_x_adjustment (GimpPreviewImage * preview_image)
{
  g_return_val_if_fail (GIMP_IS_PREVIEW_IMAGE (preview_image), NULL);
  return preview_image->x_adjustment;
}

/**
 * gimp_preview_image_get_y_adjustment:
 * @preview_image: A #GimpPreviewImage.
 * 
 * Return the adjustment that manages the y-coordinate of
 * the preview position. The value of this adjustment
 * is expressed in #preview-coordinates.
 * 
 * Applications should only change the value of the adjustment,
 * the other properties of the adjustment are managed by the widget.
 * When you try to set the value directly (i.e. not via a GtkScale widget)
 * you should use gtk_adjustment_set_correct_value() (don't use
 * gtk_adjustment_set_value() because it is buggy).
 * 
 * Return value: The #GtkAdjustment.
 **/
GtkAdjustment *
gimp_preview_image_get_y_adjustment (GimpPreviewImage * preview_image)
{
  g_return_val_if_fail (GIMP_IS_PREVIEW_IMAGE (preview_image), NULL);
  return preview_image->y_adjustment;
}

/**
 * gimp_preview_image_get_scale_adjustment:
 * @preview_image: A #GimpPreviewImage.
 * 
 * Return the adjustment that manages the scale of
 * the preview position.
 * The value of this adjustment
 * is the ratio between the size of the scaled image
 * and the size of the original image, e.g. a value
 * of 0.5 means that the preview image is reduced by
 * 50%.
 *
 * Applications should only change the value of the adjustment,
 * the other properties of the adjustment are managed by the widget.
 * When you try to set the value directly (i.e. not via a GtkScale widget)
 * you should use gtk_adjustment_set_correct_value() (don't use
 * gtk_adjustment_set_value() because it is buggy).
 * 
 * Return value: The #GtkAdjustment.
 **/
GtkAdjustment *
gimp_preview_image_get_scale_adjustment (GimpPreviewImage * preview_image)
{
  g_return_val_if_fail (GIMP_IS_PREVIEW_IMAGE (preview_image), NULL);
  return preview_image->scale_adjustment;
}

/**
 * gimp_preview_image_get_show_original:
 * @preview_image: A #GimpPreviewImage.
 * 
 * Query the @preview_image whether it shows the original image
 * during scrolls.
 * 
 * Return value: TRUE when the preview shows a scaled version
 * of the original image during scrolls.
 **/
gboolean
gimp_preview_image_get_show_original (GimpPreviewImage * preview_image)
{
  g_return_val_if_fail (GIMP_IS_PREVIEW_IMAGE (preview_image), TRUE);

  return (preview_image->show_original);
}

/**
 * gimp_preview_image_set_show_original:
 * @preview_image: A #GimpPreviewImage.
 * @show_original: TRUE when the preview shows a scaled version
 * of the original image during scrolls.
 * 
 * Set whether the @preview_image should show a scaled version
 * of the original image during scrolls.
 * When show_original is FALSE only the results of the rendering
 * function will be shown in the display.
 * This will only give acceptable results when the rendering
 * function is very fast.
 **/
void
gimp_preview_image_set_show_original (GimpPreviewImage * preview_image,
                                      gboolean show_original)
{
  g_return_if_fail (GIMP_IS_PREVIEW_IMAGE (preview_image));

  preview_image->show_original = show_original;
}
