/* GIMP Plugin Preview Widget                                                 
 * Copyright (C) 2003 Ernst Lippe
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.           
 *                                                                            
 * This library is distributed in the hope that it will be useful,             
 * but WITHOUT ANY WARRANTY; without even the implied warranty of              
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Version $Id: gimppreviewutil.c,v 1.2 2003/06/11 11:09:15 el Exp $
 */

#include "gimppreviewutil.h"

#include <string.h>

#ifndef lint
static char vcid[] G_GNUC_UNUSED =
  "$Id: gimppreviewutil.c,v 1.2 2003/06/11 11:09:15 el Exp $";
#endif /* lint */

/* Convert coordinate in preview space to image coordinates */
#define p2i(preview_coordinate, scale) ((gint)((preview_coordinate) / (scale)))

static gint
i2p (gint image_coordinate, gdouble scale)
/* Return maximum preview coordinate p such that
   p2i(p, scale) <= image_coordinate < p2i(p + 1,scale) */
{
  gint p = (gint) ((image_coordinate + 1) * scale) + 1;
  g_assert (p2i (p, scale) > image_coordinate);
  while (p2i (p, scale) > image_coordinate)
    {
      p--;
    };
  return (p);
}

/* Copy the contents of a drawable. */
static void
copy_drawable (GimpDrawable * src_drawable, GimpDrawable * dest_drawable,
               gint x, gint y, gint width, gint height)
{
  GimpPixelRgn src_rgn, dest_rgn;
  gpointer pr;

  gimp_pixel_rgn_init (&src_rgn, src_drawable, x, y, width, height, FALSE,
                       FALSE);
  gimp_pixel_rgn_init (&dest_rgn, dest_drawable, 0, 0, width, height, TRUE,
                       FALSE);
  for (pr = gimp_pixel_rgns_register (2, &src_rgn, &dest_rgn); pr != NULL;
       pr = gimp_pixel_rgns_process (pr))
    {
      gint pixelwidth = src_rgn.w * src_rgn.bpp;
      guchar *s = src_rgn.data;
      guchar *d = dest_rgn.data;
      gint h = src_rgn.h;

      while (h--)
        {
          memcpy (d, s, pixelwidth);
          s += src_rgn.rowstride;
          d += dest_rgn.rowstride;
        }
    }

  gimp_drawable_flush (dest_drawable);
}

/* Copy the contents of a drawable. */
static void
scale_drawable (GimpDrawable * src_drawable, GimpDrawable * dest_drawable,
                gdouble scale, gint preview_x, gint preview_y,
                gint preview_width, gint preview_height)
{
  GimpPixelRgn src_region, dest_region;
  guchar *image_data = NULL;
  guchar *row_buffer = NULL;

  /* Coordinates within the image of top left pixel in preview */
  const gint image_x = p2i (preview_x, scale);
  const gint image_y = p2i (preview_y, scale);


  /* Determine the available area in the drawable */
  const gint image_width =
    CLAMP (p2i (preview_x + preview_width - 1, scale) + 1 - image_x,
           0,
           src_drawable->width - image_x);

  const gint image_height =
    CLAMP (p2i (preview_y + preview_height - 1, scale) + 1 - image_y,
           0,
           src_drawable->height - image_y);

  /* This is the area in the preview that we can actually draw */
  const gint width = CLAMP (preview_width,
                            0,
                            i2p (image_x + image_width - 1,
                                 scale) + 1 - preview_x);
  const gint height = CLAMP (preview_height,
                             0,
                             i2p (image_y + image_height - 1,
                                  scale) + 1 - preview_y);

  g_assert (image_x + image_width <= src_drawable->width);
  g_assert (image_y + image_height <= src_drawable->height);

  gimp_pixel_rgn_init (&src_region, src_drawable,
                       image_x, image_y, image_width, image_height,
                       FALSE, FALSE);
  gimp_pixel_rgn_init (&dest_region, dest_drawable,
                       0, 0, preview_width, preview_height, FALSE, FALSE);

  row_buffer = g_new0 (guchar, 4 * preview_width);
  image_data =
    g_new0 (guchar, src_drawable->bpp * image_width * image_height);

  gimp_pixel_rgn_get_rect (&src_region, image_data,
                           image_x, image_y, image_width, image_height);


  /* Scale and convert to RGB */
  if (src_drawable->bpp == 4)
    {
      gint dy;
      const gint bpp = 4;

      /* For each row */
      for (dy = 0; dy < height; dy++)
        {
          /* Calculate: starting position in drawable */
          gint dx;
          /* y coordinate in image */
          const gint sy = p2i (dy + preview_y, scale);
          const guchar *const src_ptr =
            image_data + ((sy - image_y) * image_width - image_x) * bpp;

          for (dx = 0; dx < width; dx++)
            {
              /* x coordinate in image */
              const gint sx = p2i (dx + preview_x, scale);

              row_buffer[dx * 3 + 0] = src_ptr[sx * bpp + 0];
              row_buffer[dx * 3 + 1] = src_ptr[sx * bpp + 1];
              row_buffer[dx * 3 + 2] = src_ptr[sx * bpp + 2];
              row_buffer[dx * 3 + 3] = src_ptr[sx * bpp + 3];
            }
          gimp_pixel_rgn_set_row (&dest_region, row_buffer, 0, dy, width);
        }
    }
  else if (src_drawable->bpp == 3)
    {
      gint dy;
      const gint bpp = 3;

      /* For each row of the preview ... */
      for (dy = 0; dy < height; dy++)
        {
          /* Calculate: starting position in drawable */
          gint dx;
          /* y coordinate in image */
          const gint sy = p2i (dy + preview_y, scale);

          const guchar *const src_ptr =
            image_data + ((sy - image_y) * image_width - image_x) * bpp;

          for (dx = 0; dx < width; dx++)
            {
              /* x coordinate in image */
              const gint sx = p2i (dx + preview_x, scale);

              row_buffer[dx * bpp + 0] = src_ptr[sx * bpp + 0];
              row_buffer[dx * bpp + 1] = src_ptr[sx * bpp + 1];
              row_buffer[dx * bpp + 2] = src_ptr[sx * bpp + 2];
            }
          gimp_pixel_rgn_set_row (&dest_region, row_buffer, 0, dy, width);
        }

    }
  else if (src_drawable->bpp == 2)
    {
      gint dy;
      const gint bpp = 2;

      for (dy = 0; dy < height; dy++)
        {
          /* Calculate: starting position in drawable */
          gint dx;
          /* y coordinate in image */
          const gint sy = p2i (dy + preview_y, scale);

          const guchar *const src_ptr =
            image_data + ((sy - image_y) * image_width - image_x) * bpp;

          for (dx = 0; dx < width; dx++)
            {
              /* x coordinate in image */
              const gint sx = p2i (dx + preview_x, scale);

              row_buffer[dx * bpp + 0] = src_ptr[sx * bpp + 0];
              row_buffer[dx * bpp + 1] = src_ptr[sx * bpp + 1];
            }

          gimp_pixel_rgn_set_row (&dest_region, row_buffer, 0, dy, width);
        }
    }
  else if (src_drawable->bpp == 1)
    {
      gint dy;
      const gint bpp = 1;

      /* For each row of the preview ... */
      for (dy = 0; dy < height; dy++)
        {
          /* Calculate: starting position in drawable */
          gint dx;
          /* y coordinate in image */
          const gint sy = p2i (dy + preview_y, scale);

          const guchar *const src_ptr =
            image_data + ((sy - image_y) * image_width - image_x) * bpp;

          for (dx = 0; dx < width; dx++)
            {
              /* x coordinate in image */
              const gint sx = p2i (dx + preview_x, scale);

              row_buffer[dx * bpp + 0] = src_ptr[sx * bpp + 0];
            }

          gimp_pixel_rgn_set_row (&dest_region, row_buffer, 0, dy, width);
        }
    }

  gimp_drawable_flush (dest_drawable);

  g_free (image_data);
  g_free (row_buffer);
}

/* Copy the contents of a drawable. */
static void
copy_scale_drawable (GimpDrawable * src_drawable,
                     GimpDrawable * dest_drawable, gdouble scale, gint x,
                     gint y, gint width, gint height)
{
  if (scale == 1.0)
    {
      return copy_drawable (src_drawable, dest_drawable, x, y, width, height);
    }
  else
    {
      return (scale_drawable (src_drawable, dest_drawable,
                              scale, x, y, width, height));
    }
}


/* Copy a drawable that is a layer to a new image */
static GimpDrawable *
copy_layer (GimpDrawable * drawable, gint32 new_image_id,
            gdouble scale, gint x, gint y, gint width, gint height)
{
  gint32 new_layer_id = gimp_layer_new (new_image_id, "preview",
                                        width, height,
                                        gimp_drawable_type (drawable->
                                                            drawable_id),
                                        0.0, GIMP_NORMAL_MODE);
  GimpDrawable *new_drawable = gimp_drawable_get (new_layer_id);

  copy_scale_drawable (drawable, new_drawable, scale, x, y, width, height);
  gimp_image_add_layer (new_image_id, new_layer_id, 0);
  gimp_image_set_active_layer (new_image_id, new_layer_id);

  /* layer mask handling */
  if (gimp_layer_mask (drawable->drawable_id) >= 0)
    {
      gint32 old_layer_mask_id = gimp_layer_mask (drawable->drawable_id);
      GimpDrawable *old_layer_mask_drawable =
        gimp_drawable_get (old_layer_mask_id);
      gint32 new_layer_mask_id =
        gimp_layer_create_mask (new_layer_id, GIMP_ADD_WHITE_MASK);
      GimpDrawable *new_layer_mask_drawable =
        gimp_drawable_get (new_layer_mask_id);

      copy_scale_drawable (old_layer_mask_drawable, new_layer_mask_drawable,
                           scale, x, y, width, height);
      gimp_image_add_layer_mask (new_image_id, new_layer_id,
                                 new_layer_mask_id);
      gimp_layer_set_apply_mask (new_layer_id,
                                 gimp_layer_get_apply_mask (drawable->
                                                            drawable_id));
    }

  return (new_drawable);
}

/* Copy a drawable that is a channel to a new image. */
static GimpDrawable *
copy_channel (GimpDrawable * drawable, gint32 new_image_id,
              gdouble scale, gint x, gint y, gint width, gint height)
{
  GimpRGB channel_color = { 1.0, 1.0, 1.0, 1.0 };

  gint32 new_channel_id = gimp_channel_new (new_image_id, "preview",
                                            width, height,
                                            0.0, &channel_color);
  GimpDrawable *new_drawable = gimp_drawable_get (new_channel_id);

  copy_scale_drawable (drawable, new_drawable, scale, x, y, width, height);
  gimp_image_add_channel (new_image_id, new_channel_id, 0);

  gimp_image_set_active_channel (new_image_id, new_channel_id);

  return (new_drawable);
}

/* Copy the the active components from old_image_id to new_image_id */
static void
copy_active_components (gint32 old_image_id, gint32 new_image_id)
{
  GimpChannelType component;

  for (component = GIMP_RED_CHANNEL; component <= GIMP_ALPHA_CHANNEL;
       component++)
    {
      gimp_image_set_component_active (new_image_id,
                                       component,
                                       gimp_image_get_component_active
                                       (old_image_id, component));
    }
}

/* Copy the selection from old_image_id to new_image_id */
static void
copy_selection (gint32 old_image_id, gint32 new_image_id,
                gdouble scale, gint x, gint y, gint width, gint height)
{

  if (!gimp_selection_is_empty (old_image_id))
    {
      gint32 old_selection_channel_id =
        gimp_image_get_selection (old_image_id);
      gint32 new_selection_channel_id =
        gimp_image_get_selection (new_image_id);
      GimpDrawable *old_selection_drawable =
        gimp_drawable_get (old_selection_channel_id);
      GimpDrawable *new_selection_drawable =
        gimp_drawable_get (new_selection_channel_id);

      /* Copy the selection */
      gimp_selection_none (new_image_id);
      copy_scale_drawable (old_selection_drawable, new_selection_drawable,
                           scale, x, y, width, height);
      /* This is a hack otherwise the GIMP does not know that
         the selection has changed */
      gimp_selection_load (new_selection_channel_id);

      if (gimp_selection_is_empty (new_image_id))
        {
          /* This will happens when the selected area is outside the maskbounds.
             Actually the whole image is not selected, but the GIMP treats this
             as though the whole image has been selected.
             This is solved by selecting a point that is outside the drawable.
           */
          gimp_rect_select (new_image_id, width, height, 1, 1,
                            GIMP_CHANNEL_OP_REPLACE, FALSE, 0.0);
        }
    }
  else
    {
      /* Without this call layers are updated correctly, but
         channels are not. This might be a bug in the GIMP.
       */
      gimp_selection_all (new_image_id);
    }
}

/**
 * gimp_preview_get_scaled_temp_drawable:
 * @drawable: The original #GimpDrawable.
 * @scale: Scaling factor.
 * @x: Left coordinate of the area in #preview-coordinates.
 * @y: Top coordinate of the area in #preview-coordinates.
 * @width: Width of the area in #preview-coordinates.
 * @height: Height of the area in #preview-coordinates.
 * 
 * Make a new #GimpDrawable as a scaled copy of an area from
 * another #GimpDrawable.
 * This function creates a new #GimpImage to hold the result.
 * This function also copies the selection and active components.
 * If the @drawable is a layer that has a layer mask it is copied too.
 * When the result is no longer needed it must be freed with gimp_preview_free_temp_drawable().
 * For performance reasons undo's have been disabled on the image.
 *
 * Return value: The new #GimpDrawable with a size of @width x @height.
 **/
GimpDrawable *
gimp_preview_get_scaled_temp_drawable (GimpDrawable * drawable, gdouble scale,
                                       gint x, gint y, gint width,
                                       gint height)
{
  gint32 old_image_id = gimp_drawable_image (drawable->drawable_id);
  /* The image is one pixel bigger than the drawable, see copy_selection(). */
  gint32 new_image_id = gimp_image_new (width + 1, height + 1,
                                        gimp_image_base_type (old_image_id));

  GimpDrawable *new_drawable = NULL;

  gimp_image_undo_disable (new_image_id);

  copy_selection (old_image_id, new_image_id, scale, x, y, width, height);
  copy_active_components (old_image_id, new_image_id);

  if (gimp_drawable_is_layer (drawable->drawable_id) ||
      gimp_drawable_is_layer_mask (drawable->drawable_id))
    {
      new_drawable = copy_layer (drawable, new_image_id,
                                 scale, x, y, width, height);

    }
  else if (gimp_drawable_is_channel (drawable->drawable_id))
    {
      new_drawable = copy_channel (drawable, new_image_id,
                                   scale, x, y, width, height);
    }
  else
    {
      g_assert ("gimp_preview_get_temp_drawable: Unknown drawable type");
    }


  return (new_drawable);
}

/**
 * gimp_preview_get_temp_drawable:
 * @drawable: The original #GimpDrawable.
 * @x: Left coordinate of the area.
 * @y: Top coordinate of the area.
 * @width: Width of the area.
 * @height: Height of the area.
 * 
 * Make a new #GimpDrawable as a copy of an area from another #GimpDrawable.
 * This function creates a new #GimpImage to hold the result.
 * This function also copies the selection and active components.
 * If the @drawable is a layer that has a layer mask it is copied too.
 * When the result is no longer needed it must be freed with gimp_preview_free_temp_drawable().
 * For performance reasons undo's have been disabled on the image.
 *
 * Return value: The new #GimpDrawable.
 **/
GimpDrawable *
gimp_preview_get_temp_drawable (GimpDrawable * drawable, gint x, gint y,
                                gint width, gint height)
{
  return (gimp_preview_get_scaled_temp_drawable
          (drawable, 1.0, x, y, width, height));
}

/**
 * gimp_preview_free_temp_drawable:
 * @drawable: A #GimpDrawable that was created with gimp_preview_get_temp_drawable().
 * 
 * Free a drawable that has been created with gimp_preview_get_temp_drawable()
 * or gimp_preview_get_scaled_temp_drawable().
 **/
void
gimp_preview_free_temp_drawable (GimpDrawable * drawable)
{
  gint32 image_id = gimp_drawable_image (drawable->drawable_id);

  gimp_image_delete (image_id);
}
