/* The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * Pixelize plug-in (ported to GIMP v1.0)
 * Copyright (C) 1997 Eiichi Takamori <taka@ma1.seikyou.ne.jp>
 * original pixelize.c for GIMP 0.54 by Tracy Scott
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* 
 * This version was hacked to work with my version of the preview.
 * Known bugs:
 * Changing the pixelsize does not work when the width and height
 * are "chained". This appears to be a bug in
 * gimp_coordinates_callback that stops emission of the value-changed
 * signal.
 * The preview does not function well with selections. The area
 * outside the selection is black when the layer has no alpha
 * or transparent when the layer has alpha. This could be solved
 * by initializing the dest_drawable with a copy of the corresponding
 * area in the src_drawable.
 * The progress is not correctly updated for pixelize_small.
 * Ernst Lippe 2003-03-04
 */


/*
 * version 1.04
 * This version requires GIMP v0.99.10 or above.
 *
 * This plug-in "pixelizes" the image.
 *
 *	Eiichi Takamori <taka@ma1.seikyou.ne.jp>
 *	http://ha1.seikyou.ne.jp/home/taka/gimp/
 *
 * Changes from version 1.03 to version 1.04:
 * - Added gtk_rc_parse
 * - Added entry with scale
 * - Fixed bug that large pixelwidth >=64 sometimes caused core dump
 *
 * Changes from gimp-0.99.9/plug-ins/pixelize.c to version 1.03:
 * - Fixed comments and help strings
 * - Fixed `RGB, GRAY' to `RGB*, GRAY*'
 * - Fixed procedure db name `pixelize' to `plug_in_pixelize'
 *
 * Differences from Tracy Scott's original `pixelize' plug-in:
 *
 * - Algorithm is modified to work around with the tile management.
 *   The way of pixelizing is switched by the value of pixelwidth.  If
 *   pixelwidth is greater than (or equal to) tile width, then this
 *   plug-in makes GimpPixelRgn with that width and proceeds. Otherwise,
 *   it makes the region named `PixelArea', whose size is smaller than
 *   tile width and is multiply of pixel width, and acts onto it.
 */

/* pixelize filter written for the GIMP
 *  -Tracy Scott
 *
 * This filter acts as a low pass filter on the color components of
 * the provided region
 */

/* #include "config.h" */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#include "gimppreviewdialog.h"
#include "gimppreviewprogressupdater.h"
#include "gimppreview.h"
#include "gimppreviewutil.h"

#ifndef lint
static char vcid[] G_GNUC_UNUSED =
  "$Id: pixelize.c,v 1.5 2003/06/11 11:09:17 el Exp $";
#endif /* lint */

#define _(x) (x)
#define N_(x) (x)

#ifdef RCSID
static char rcsid[] = "$Id: pixelize.c,v 1.5 2003/06/11 11:09:17 el Exp $";
#endif


/* Some useful macros */

#define TILE_CACHE_SIZE  16
#define SCALE_WIDTH     125
#define ENTRY_WIDTH      50

typedef struct
{
  gint pixelwidth;
  gint pixelheight;
}
PixelizeValues;

typedef struct
{
  GtkWidget *sizeentry;
  gint run;
}
PixelizeInterface;

typedef struct
{
  gint x, y, w, h;
  gint width;
  gint height;
  guchar *data;
}
PixelArea;

/* Declare local functions.
 */
static void query (void);
static void run (gchar * name,
                 gint nparams,
                 GimpParam * param,
                 gint * nreturn_vals, GimpParam ** return_vals);

static gint pixelize_dialog (GimpDrawable * drawable);
static void pixel_size_update_callback (GtkWidget * widget, gpointer data);

static void pixelize (GimpDrawable * src_drawable,
                      GimpDrawable * dest_drawable, gint x1, gint y1,
                      gint width, gint height,
                      GimpPreviewProgressUpdater * progress_updater);
static void pixelize_large (GimpDrawable * src_drawable,
                            GimpDrawable * dest_drawable, gint pixelwidth,
                            gint pixelheight, gint x1, gint y1, gint width,
                            gint height,
                            GimpPreviewProgressUpdater * progress_updater);
static void pixelize_small (GimpDrawable * src_drawable,
                            GimpDrawable * dest_drawable, gint pixelwidth,
                            gint pixelheight, gint tile_width,
                            gint tile_height, gint x1, gint y1, gint width,
                            gint height,
                            GimpPreviewProgressUpdater * progress_updater);
static void pixelize_sub (gint pixelwidth, gint pixelheight, gint bpp,
                          PixelArea * area);
static void pixelize_callback (GtkWidget * widget,
                               GimpPreviewUpdateEvent * event, gpointer data);

/***** Local vars *****/

GimpPlugInInfo PLUG_IN_INFO = {
  NULL,                         /* init_proc  */
  NULL,                         /* quit_proc  */
  query,                        /* query_proc */
  run                           /* run_proc   */
};

static PixelizeValues pvals = {
  10,
  10
};

static PixelizeInterface pint = {
  NULL,
  FALSE                         /* run */
};

/***** Functions *****/

MAIN ()
     static void query (void)
{
  static GimpParamDef pixelize_args[] = {
    {GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive"},
    {GIMP_PDB_IMAGE, "image", "Input image (unused)"},
    {GIMP_PDB_DRAWABLE, "drawable", "Input drawable"},
    {GIMP_PDB_INT32, "pixelwidth", "Pixel width (the decrease in resolution)"}
  };
  static GimpParamDef pixelize2_args[] = {
    {GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive"},
    {GIMP_PDB_IMAGE, "image", "Input image (unused)"},
    {GIMP_PDB_DRAWABLE, "drawable", "Input drawable"},
    {GIMP_PDB_INT32, "pixelwidth",
     "Pixel width (the decrease in horizontal resolution)"},
    {GIMP_PDB_INT32, "pixelheight",
     "Pixel height (the decrease in vertical resolution)"}
  };

  gimp_install_procedure ("plug_in_pixelize",
                          "Pixelize the contents of the specified drawable",
                          "Pixelize the contents of the specified drawable with speficied pixelizing width.",
                          "Spencer Kimball & Peter Mattis, Tracy Scott, (ported to 1.0 by) Eiichi Takamori",
                          "Spencer Kimball & Peter Mattis, Tracy Scott",
                          "1995",
                          /* N_("<Image>/Filters/Blur/Pixelize..."), */
                          "<Image>/Filters/Generic/Preview Pixelize...",
                          "RGB*, GRAY*",
                          GIMP_PLUGIN,
                          G_N_ELEMENTS (pixelize_args), 0,
                          pixelize_args, NULL);

  gimp_install_procedure ("plug_in_pixelize2",
                          "Pixelize the contents of the specified drawable",
                          "Pixelize the contents of the specified drawable with speficied pixelizing width.",
                          "Spencer Kimball & Peter Mattis, Tracy Scott, (ported to 1.0 by) Eiichi Takamori",
                          "Spencer Kimball & Peter Mattis, Tracy Scott",
                          "2001",
                          NULL,
                          "RGB*, GRAY*",
                          GIMP_PLUGIN,
                          G_N_ELEMENTS (pixelize2_args), 0,
                          pixelize2_args, NULL);
}

static void
run (gchar * name,
     gint nparams,
     GimpParam * param, gint * nreturn_vals, GimpParam ** return_vals)
{
  static GimpParam values[1];
  GimpDrawable *drawable;
  GimpRunMode run_mode;
  GimpPDBStatusType status = GIMP_PDB_SUCCESS;

  run_mode = param[0].data.d_int32;

  *nreturn_vals = 1;
  *return_vals = values;

  values[0].type = GIMP_PDB_STATUS;
  values[0].data.d_status = status;

  /*  Get the specified drawable  */
  drawable = gimp_drawable_get (param[2].data.d_drawable);

  switch (run_mode)
    {
    case GIMP_RUN_INTERACTIVE:
      /*  Possibly retrieve data  */
      gimp_get_data ("plug_in_pixelize", &pvals);

      /*  First acquire information with a dialog  */
      if (!pixelize_dialog (drawable))
        {
          gimp_drawable_detach (drawable);
          return;
        }
      break;

    case GIMP_RUN_NONINTERACTIVE:
      /*  Make sure all the arguments are there!  */
      if ((!strcmp (name, "plug_in_pixelize") && nparams != 4) ||
          (!strcmp (name, "plug_in_pixelize2") && nparams != 5))
        {
          status = GIMP_PDB_CALLING_ERROR;
        }

      if (status == GIMP_PDB_SUCCESS)
        {
          pvals.pixelwidth = (gdouble) param[3].data.d_int32;

          if (nparams == 4)
            pvals.pixelheight = pvals.pixelwidth;
          else
            pvals.pixelheight = (gdouble) param[4].data.d_int32;
        }

      if ((status == GIMP_PDB_SUCCESS) &&
          (pvals.pixelwidth <= 0 || pvals.pixelheight <= 0))
        {
          status = GIMP_PDB_CALLING_ERROR;
        }
      break;

    case GIMP_RUN_WITH_LAST_VALS:
      /*  Possibly retrieve data  */
      gimp_get_data ("plug_in_pixelize", &pvals);
      break;

    default:
      break;
    }

  if (status == GIMP_PDB_SUCCESS)
    {
      /*  Make sure that the drawable is gray or RGB color  */
      if (gimp_drawable_is_rgb (drawable->drawable_id) ||
          gimp_drawable_is_gray (drawable->drawable_id))
        {
          gint x, x2, y, y2, width, height;
          GimpPreviewProgressUpdater gimp_progress_updater;

          gimp_drawable_mask_bounds (drawable->drawable_id, &x, &y, &x2, &y2);
          width = x2 - x;
          height = y2 - y;
          gimp_progress_init (_("Pixelizing..."));
          gimp_preview_progress_updater_init_for_gimp
            (&gimp_progress_updater);

          /*  set the tile cache size  */
          gimp_tile_cache_ntiles (TILE_CACHE_SIZE);

          /*  run the pixelize effect  */
          pixelize (drawable, drawable, x, y, width, height,
                    &gimp_progress_updater);

          if (run_mode != GIMP_RUN_NONINTERACTIVE)
            gimp_displays_flush ();

          /*  Store data  */
          if (run_mode == GIMP_RUN_INTERACTIVE)
            gimp_set_data ("plug_in_pixelize", &pvals,
                           sizeof (PixelizeValues));
        }
      else
        {
          /* g_message ("pixelize: cannot operate on indexed color images"); */
          status = GIMP_PDB_EXECUTION_ERROR;
        }
    }

  values[0].data.d_status = status;

  gimp_drawable_detach (drawable);
}

static GtkWidget *
pixelize_settings (GtkWidget * parent, gpointer data)
{
  GimpDrawable *drawable = (GimpDrawable *) data;
  GtkWidget *vbox;
  guint32 image_id;
  GimpUnit unit;
  gdouble xres, yres;

  vbox = gtk_vbox_new (FALSE, 4);
  gtk_container_set_border_width (GTK_CONTAINER (vbox), 4);

  image_id = gimp_drawable_image (drawable->drawable_id);
  unit = gimp_image_get_unit (image_id);
  gimp_image_get_resolution (image_id, &xres, &yres);

  pint.sizeentry = gimp_coordinates_new (unit, "%a", TRUE, TRUE, ENTRY_WIDTH,
                                         GIMP_SIZE_ENTRY_UPDATE_SIZE,
                                         FALSE, FALSE,
                                         _("Pixel _Width:"),
                                         pvals.pixelwidth, xres,
                                         1, drawable->width,
                                         1, drawable->width,
                                         _("Pixel _Height:"),
                                         pvals.pixelheight, yres,
                                         1, drawable->height,
                                         1, drawable->height);

  g_signal_connect (G_OBJECT (pint.sizeentry),
                    "value-changed",
                    G_CALLBACK (pixel_size_update_callback), NULL);

  gtk_box_pack_start (GTK_BOX (vbox), pint.sizeentry, FALSE, FALSE, 0);
  gtk_widget_show (pint.sizeentry);

  return vbox;
}

static gboolean
pixelize_dialog (GimpDrawable * drawable)
{
  EffectParam_t param = {
    _("Pixelize"),
    "pixelize",
    "filters/pixelize.html",
    pixelize_settings,
    drawable,
    pixelize_callback
  };

  return do_preview_dialog (drawable, &param);
}

static void
pixel_size_update_callback (GtkWidget * widget, gpointer data)
{
  printf ("Update\n");
  pvals.pixelwidth =
    gimp_size_entry_get_refval (GIMP_SIZE_ENTRY (pint.sizeentry), 0);

  pvals.pixelheight =
    gimp_size_entry_get_refval (GIMP_SIZE_ENTRY (pint.sizeentry), 1);

  preview_update_effect ();
}

/*
  Pixelize Effect
 */

static void
pixelize (GimpDrawable * src_drawable, GimpDrawable * dest_drawable,
          gint x1, gint y1, gint width, gint height,
          GimpPreviewProgressUpdater * progress_updater)
{
  gint tile_width;
  gint tile_height;
  gint pixelwidth;
  gint pixelheight;

  tile_width = gimp_tile_width ();
  tile_height = gimp_tile_height ();

  pixelwidth = pvals.pixelwidth;
  pixelheight = pvals.pixelheight;

  if (pixelwidth < 0)
    pixelwidth = -pixelwidth;
  if (pixelwidth < 1)
    pixelwidth = 1;

  if (pixelheight < 0)
    pixelheight = -pixelheight;
  if (pixelheight < 1)
    pixelheight = 1;

  if (pixelwidth >= tile_width || pixelheight >= tile_height)
    pixelize_large (src_drawable, dest_drawable, pixelwidth, pixelheight, x1,
                    y1, width, height, progress_updater);
  else
    pixelize_small (src_drawable, dest_drawable, pixelwidth, pixelheight,
                    tile_width, tile_height, x1, y1, width, height,
                    progress_updater);
}

/*
  This function operates on the image when pixelwidth >= tile_width.
  It simply sets the size of GimpPixelRgn as pixelwidth and proceeds.
 */
static void
pixelize_large (GimpDrawable * src_drawable,
                GimpDrawable * dest_drawable,
                gint pixelwidth,
                gint pixelheight,
                gint x1,
                gint y1,
                gint width, gint height,
                GimpPreviewProgressUpdater * progress_updater)
{
  GimpPixelRgn src_rgn, dest_rgn;
  guchar *src_row, *dest_row;
  guchar *src, *dest;
  gulong *average;
  gint row, col, b, bpp;
  gint x, y, x_step, y_step;
  gulong count;
  gint x2, y2;
  gint progress, max_progress;
  gpointer pr;


  x2 = x1 + width;
  y2 = y1 + height;

  bpp = gimp_drawable_bpp (src_drawable->drawable_id);
  average = g_new (gulong, bpp);

  /* Initialize progress */
  progress = 0;
  max_progress = 2 * (x2 - x1) * (y2 - y1);

  for (y = y1; y < y2; y += pixelheight - (y % pixelheight))
    {
      for (x = x1; x < x2; x += pixelwidth - (x % pixelwidth))
        {
          x_step = pixelwidth - (x % pixelwidth);
          y_step = pixelheight - (y % pixelheight);
          x_step = MIN (x_step, x2 - x);
          y_step = MIN (y_step, y2 - y);

          gimp_pixel_rgn_init (&src_rgn, src_drawable,
                               x, y, x_step, y_step, FALSE, FALSE);
          for (b = 0; b < bpp; b++)
            average[b] = 0;
          count = 0;

          for (pr = gimp_pixel_rgns_register (1, &src_rgn);
               pr != NULL; pr = gimp_pixel_rgns_process (pr))
            {
              src_row = src_rgn.data;
              for (row = 0; row < src_rgn.h; row++)
                {
                  src = src_row;
                  for (col = 0; col < src_rgn.w; col++)
                    {
                      for (b = 0; b < bpp; b++)
                        average[b] += src[b];
                      src += src_rgn.bpp;
                      count += 1;
                    }
                  src_row += src_rgn.rowstride;
                }
              /* Update progress */
              progress += src_rgn.w * src_rgn.h;
              if (!gimp_preview_progress_updater_set_fraction
                  (progress_updater,
                   (gdouble) progress / (gdouble) max_progress))
                return;
            }

          if (count > 0)
            {
              for (b = 0; b < bpp; b++)
                average[b] = (guchar) (average[b] / count);
            }

          if (src_drawable == dest_drawable)
            {
              gimp_pixel_rgn_init (&dest_rgn, dest_drawable,
                                   x, y, x_step, y_step, TRUE, TRUE);
            }
          else
            {
              gimp_pixel_rgn_init (&dest_rgn, dest_drawable,
                                   x - x1, y - y1, x_step, y_step, TRUE,
                                   TRUE);
            }
          for (pr = gimp_pixel_rgns_register (1, &dest_rgn);
               pr != NULL; pr = gimp_pixel_rgns_process (pr))
            {
              dest_row = dest_rgn.data;
              for (row = 0; row < dest_rgn.h; row++)
                {
                  dest = dest_row;
                  for (col = 0; col < dest_rgn.w; col++)
                    {
                      for (b = 0; b < bpp; b++)
                        dest[b] = average[b];
                      dest += dest_rgn.bpp;
                      count += 1;
                    }
                  dest_row += dest_rgn.rowstride;
                }
              /* Update progress */
              progress += dest_rgn.w * dest_rgn.h;
              if (!gimp_preview_progress_updater_set_fraction
                  (progress_updater,
                   (gdouble) progress / (gdouble) max_progress))
                return;
            }
        }
    }

  g_free (average);

  /*  update the blurred region  */
  gimp_drawable_flush (dest_drawable);
  gimp_drawable_merge_shadow (dest_drawable->drawable_id, TRUE);
  gimp_drawable_update (dest_drawable->drawable_id, x1, y1, (x2 - x1),
                        (y2 - y1));
}


/*
   This function operates on PixelArea, whose width and height are
   multiply of pixel width, and less than the tile size (to enhance
   its speed).

   If any coordinates of mask boundary is not multiply of pixel width
   (e.g.  x1 % pixelwidth != 0), operates on the region whose width
   or height is the remainder.
 */
static void
pixelize_small (GimpDrawable * src_drawable,
                GimpDrawable * dest_drawable,
                gint pixelwidth,
                gint pixelheight,
                gint tile_width,
                gint tile_height,
                gint x1,
                gint y1,
                gint width, gint height,
                GimpPreviewProgressUpdater * progress_updater)
{
  GimpPixelRgn src_rgn, dest_rgn;
  gint bpp;
  gint x2, y2;
  gint progress, max_progress;
  gint offset_x, offset_y;
  PixelArea area;

  x2 = x1 + width;
  y2 = y1 + height;
  gimp_pixel_rgn_init (&src_rgn, src_drawable, x1, y1, x2 - x1, y2 - y1,
                       FALSE, FALSE);
  if (src_drawable == dest_drawable)
    {
      offset_x = 0;
      offset_y = 0;
    }
  else
    {
      offset_x = x1;
      offset_y = y1;
    }

  gimp_pixel_rgn_init (&dest_rgn, dest_drawable, x1 - offset_x, y1 - offset_y,
                       x2 - x1, y2 - y1, TRUE, TRUE);

  /* Initialize progress */
  progress = 0;
  max_progress = (x2 - x1) * (y2 - y1);

  bpp = src_drawable->bpp;

  area.width = (tile_width / pixelwidth) * pixelwidth;
  area.height = (tile_height / pixelheight) * pixelheight;
  area.data = g_new (guchar, (glong) bpp * area.width * area.height);

  for (area.y = y1; area.y < y2;
       area.y += area.height - (area.y % area.height))
    {
      area.h = area.height - (area.y % area.height);
      area.h = MIN (area.h, y2 - area.y);

      for (area.x = x1; area.x < x2;
           area.x += area.width - (area.x % area.width))
        {
          area.w = area.width - (area.x % area.width);
          area.w = MIN (area.w, x2 - area.x);

          gimp_pixel_rgn_get_rect (&src_rgn, area.data,
                                   area.x, area.y, area.w, area.h);

          pixelize_sub (pixelwidth, pixelheight, bpp, &area);

          gimp_pixel_rgn_set_rect (&dest_rgn, area.data,
                                   area.x - offset_x, area.y - offset_y,
                                   area.w, area.h);

          /* Update progress */
          progress += area.w * area.h;
          if (!gimp_preview_progress_updater_set_fraction (progress_updater,
                                                           (gdouble) progress
                                                           /
                                                           (gdouble)
                                                           max_progress))
            return;
        }
    }

  g_free (area.data);

  /*  update the pixelized region  */
  gimp_drawable_flush (dest_drawable);
  gimp_drawable_merge_shadow (dest_drawable->drawable_id, TRUE);
  gimp_drawable_update (dest_drawable->drawable_id, x1, y1, (x2 - x1),
                        (y2 - y1));
}

/*
  This function acts on one PixelArea.	Since there were so many
  nested FORs in pixelize_small(), I put a few of them here...
  */

static void
pixelize_sub (gint pixelwidth, gint pixelheight, gint bpp, PixelArea * area)
{
  glong average[4];             /* bpp <= 4 */
  gint x, y, w, h;
  guchar *buf_row, *buf;
  gint row, col;
  gint rowstride;
  gint count;
  gint i;

  rowstride = area->w * bpp;

  for (y = area->y; y < area->y + area->h;
       y += pixelheight - (y % pixelheight))
    {
      h = pixelheight - (y % pixelheight);
      h = MIN (h, area->y + area->h - y);

      for (x = area->x; x < area->x + area->w;
           x += pixelwidth - (x % pixelwidth))
        {
          w = pixelwidth - (x % pixelwidth);
          w = MIN (w, area->x + area->w - x);

          for (i = 0; i < bpp; i++)
            average[i] = 0;
          count = 0;

          /* Read */
          buf_row =
            area->data + (y - area->y) * rowstride + (x - area->x) * bpp;

          for (row = 0; row < h; row++)
            {
              buf = buf_row;
              for (col = 0; col < w; col++)
                {
                  for (i = 0; i < bpp; i++)
                    average[i] += buf[i];
                  count++;
                  buf += bpp;
                }
              buf_row += rowstride;
            }

          /* Average */
          if (count > 0)
            {
              for (i = 0; i < bpp; i++)
                average[i] /= count;
            }

          /* Write */
          buf_row =
            area->data + (y - area->y) * rowstride + (x - area->x) * bpp;

          for (row = 0; row < h; row++)
            {
              buf = buf_row;
              for (col = 0; col < w; col++)
                {
                  for (i = 0; i < bpp; i++)
                    buf[i] = average[i];
                  count++;
                  buf += bpp;
                }
              buf_row += rowstride;
            }
        }
    }
}

static void
pixelize_callback (GtkWidget * widget, GimpPreviewUpdateEvent * event,
                   gpointer data)
{
  GimpPreview *preview = GIMP_PREVIEW (widget);
  GimpDrawable *drawable = (GimpDrawable *) data;
  GimpDrawable *dest_drawable;
  GimpPreviewProgressUpdater preview_progress_updater;

  dest_drawable =
    gimp_preview_get_temp_drawable (drawable, event->image_x, event->image_y,
                                    event->image_width, event->image_height);

  gimp_preview_progress_updater_init_for_preview (&preview_progress_updater,
                                                  preview);
  pixelize (drawable, dest_drawable, event->image_x, event->image_y,
            event->image_width, event->image_height,
            &preview_progress_updater);
  gimp_preview_draw_unscaled_drawable (preview, dest_drawable);
  gimp_preview_free_temp_drawable (dest_drawable);
}
