/* Spread --- image filter plug-in for The Gimp image manipulation program
 * Copyright (C) 1997 Brian Degenhardt and Federico Mena Quintero
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Please direct all comments, questions, bug reports  etc to Brian Degenhardt
 * bdegenha@ucsd.edu
 *
 * You can contact Federico Mena Quintero at quartic@polloux.fciencias.unam.mx
 * You can contact the original The Gimp authors at gimp@xcf.berkeley.edu
 */
/* #include "config.h" */

/* 
 * This version was hacked to work with my version of the preview.
 * Known bugs:
 * Changing the pixelsize does not work when the width and height
 * are "chained". This appears to be a bug in
 * gimp_coordinates_callback that stops emission of the value-changed
 * signal.
 * The preview does not function well with selections. The area
 * outside the selection shows garbage. This could be solved
 * by initializing the dest_drawable with a copy of the corresponding
 * area in the src_drawable.
 * Ernst Lippe 2003-03-04
 */

#include <time.h>
#include <stdio.h>
#include <stdlib.h>

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

/* #include "libgimp/stdplugins-intl.h" */

#include "gimppreviewdialog.h"
#include "gimppreviewprogressupdater.h"
#include "gimppreview.h"
#include "gimppreviewutil.h"

#ifndef lint
static char vcid[] G_GNUC_UNUSED =
  "$Id: spread.c,v 1.6 2003/06/11 11:09:18 el Exp $";
#endif /* lint */

#define _(x) (x)
#define N_(x) (x)

/* Some useful macros */
#define SCALE_WIDTH 200
#define TILE_CACHE_SIZE 16
#define ENTRY_WIDTH 50

typedef struct
{
  gdouble spread_amount_x;
  gdouble spread_amount_y;
}
SpreadValues;

typedef struct
{
  GtkWidget *size;
  gint run;
}
SpreadInterface;


/* Declare local functions.
 */
static void query (void);
static void run (gchar * name,
                 gint nparams,
                 GimpParam * param,
                 gint * nreturn_vals, GimpParam ** return_vals);

static void spread (GimpDrawable * src_drawable, GimpDrawable * dest_drawable,
                    gint x1, gint y1, gint a_width, gint a_height,
                    GimpPreviewProgressUpdater * progress_updater);

static gint spread_dialog (gint32 image_ID, GimpDrawable * drawable);
static void spread_size_update_callback (GtkWidget * widget, gpointer data);
static void spread_callback (GtkWidget * widget,
                             GimpPreviewUpdateEvent * event, gpointer data);

static GimpTile *spread_pixel (GimpDrawable * drawable,
                               GimpTile * tile,
                               gint x,
                               gint y,
                               gint * row, gint * col, guchar * pixel);


/***** Local vars *****/

GimpPlugInInfo PLUG_IN_INFO = {
  NULL,                         /* init_proc  */
  NULL,                         /* quit_proc  */
  query,                        /* query_proc */
  run,                          /* run_proc   */
};

static SpreadValues spvals = {
  5,                            /*  horizontal spread amount  */
  5                             /*  vertical spread amount    */
};

static SpreadInterface pint = {
  FALSE                         /*  run  */
};

/***** Functions *****/

MAIN ()
     static void query (void)
{
  static GimpParamDef args[] = {
    {GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive"},
    {GIMP_PDB_IMAGE, "image", "Input image (unused)"},
    {GIMP_PDB_DRAWABLE, "drawable", "Input drawable"},
    {GIMP_PDB_FLOAT, "spread_amount_x",
     "Horizontal spread amount (0 <= spread_amount_x <= 200)"},
    {GIMP_PDB_FLOAT, "spread_amount_y",
     "Vertical spread amount (0 <= spread_amount_y <= 200)"}
  };

  gimp_install_procedure ("plug_in_spread2",
                          "Spread the contents of the specified drawable",
                          "Spreads the pixels of the specified drawable.  "
                          "Pixels are randomly moved to another location whose "
                          "distance varies from the original by the horizontal "
                          "and vertical spread amounts ",
                          "Spencer Kimball and Peter Mattis, ported by Brian "
                          "Degenhardt and Federico Mena Quintero",
                          "Federico Mena Quintero and Brian Degenhardt",
                          "1997",
/*			  N_("<Image>/Filters/Noise/Spread..."), */
                          "<Image>/Filters/Generic/Preview Spread...",
                          "RGB*, GRAY*",
                          GIMP_PLUGIN, G_N_ELEMENTS (args), 0, args, NULL);
}

static void
run (gchar * name,
     gint nparams,
     GimpParam * param, gint * nreturn_vals, GimpParam ** return_vals)
{
  static GimpParam values[1];
  gint32 image_ID;
  GimpDrawable *drawable;
  GimpRunMode run_mode;
  GimpPDBStatusType status = GIMP_PDB_SUCCESS;

  run_mode = param[0].data.d_int32;

  /*  Get the specified image and drawable  */
  image_ID = param[1].data.d_image;
  drawable = gimp_drawable_get (param[2].data.d_drawable);

  *nreturn_vals = 1;
  *return_vals = values;

  values[0].type = GIMP_PDB_STATUS;
  values[0].data.d_status = status;

  switch (run_mode)
    {
    case GIMP_RUN_INTERACTIVE:
      /*  Possibly retrieve data  */
      gimp_get_data ("plug_in_spread", &spvals);

      /*  set the tile cache size  */
      gimp_tile_cache_ntiles (TILE_CACHE_SIZE);

      /*  First acquire information with a dialog  */
      if (!spread_dialog (image_ID, drawable))
        return;
      break;

    case GIMP_RUN_NONINTERACTIVE:
      /*  Make sure all the arguments are there!  */
      if (nparams != 5)
        {
          status = GIMP_PDB_CALLING_ERROR;
        }
      else
        {
          spvals.spread_amount_x = param[3].data.d_float;
          spvals.spread_amount_y = param[4].data.d_float;
        }

      if ((status == GIMP_PDB_SUCCESS) &&
          (spvals.spread_amount_x < 0 || spvals.spread_amount_x > 200) &&
          (spvals.spread_amount_y < 0 || spvals.spread_amount_y > 200))
        status = GIMP_PDB_CALLING_ERROR;
      break;

    case GIMP_RUN_WITH_LAST_VALS:
      /*  Possibly retrieve data  */
      gimp_get_data ("plug_in_spread", &spvals);
      break;

    default:
      break;
    }

  if (status == GIMP_PDB_SUCCESS)
    {
      /*  Make sure that the drawable is gray or RGB color  */
      if (gimp_drawable_is_rgb (drawable->drawable_id) ||
          gimp_drawable_is_gray (drawable->drawable_id))
        {
          gint x, x2, y, y2, width, height;
          GimpPreviewProgressUpdater gimp_progress_updater;

          gimp_drawable_mask_bounds (drawable->drawable_id, &x, &y, &x2, &y2);
          width = x2 - x;
          height = y2 - y;
          gimp_progress_init (_("Spreading..."));

          gimp_preview_progress_updater_init_for_gimp
            (&gimp_progress_updater);

          spread (drawable, drawable, x, y, width, height,
                  &gimp_progress_updater);

          if (run_mode != GIMP_RUN_NONINTERACTIVE)
            gimp_displays_flush ();

          /*  Store data  */
          if (run_mode == GIMP_RUN_INTERACTIVE)
            gimp_set_data ("plug_in_spread", &spvals, sizeof (SpreadValues));
        }
      else
        {
          /* gimp_message ("spread: cannot operate on indexed color images"); */
          status = GIMP_PDB_EXECUTION_ERROR;
        }
    }

  values[0].data.d_status = status;

  gimp_drawable_detach (drawable);
}

static GimpTile *
spread_pixel (GimpDrawable * src_drawable,
              GimpTile * src_tile,
              gint x, gint y, gint * row, gint * col, guchar * pixel)
{
  guchar *data;
  gint b;

  /* EL: Removed check if (x,y) was inside the area (x1, y1, x2, y2)
   * This caused black pixels in the preview.
   * It is only relevant that (x,y) lies within the drawable
   * and that is already checked before calling this function.
   */
  if ((x >> 6 != *col) || (y >> 6 != *row))
    {
      *col = x >> 6;
      *row = y >> 6;
      if (src_tile)
        gimp_tile_unref (src_tile, FALSE);
      src_tile = gimp_drawable_get_tile (src_drawable, FALSE, *row, *col);
      gimp_tile_ref (src_tile);
    }

  data =
    src_tile->data + src_tile->bpp * (src_tile->ewidth * (y % 64) + (x % 64));

  for (b = 0; b < src_drawable->bpp; b++)
    pixel[b] = data[b];

  return src_tile;
}

static void
spread (GimpDrawable * src_drawable, GimpDrawable * dest_drawable,
        gint x1, gint y1, gint a_width, gint a_height,
        GimpPreviewProgressUpdater * progress_updater)
{
  GimpPixelRgn dest_rgn;
  GimpTile *src_tile = NULL;
  gint row = -1;
  gint col = -1;
  gpointer pr;

  gint width, height;
  gint bytes;
  guchar *destrow;
  guchar *dest;
  guchar pixel[4][4];
  gint x2, y2;
  gint x, y;
  gint progress, max_progress;

  gdouble x_amount, y_amount;
  gdouble angle;

  gint xdist, ydist;
  gint xi, yi;

  gint k;
  gint x_mod_value, x_sub_value;
  gint y_mod_value, y_sub_value;
  gint angle_mod_value, angle_sub_value;
  gint offset_x, offset_y;

  x2 = x1 + a_width;
  y2 = y1 + a_height;

  width = src_drawable->width;
  height = src_drawable->height;
  bytes = src_drawable->bpp;

  progress = 0;
  max_progress = (x2 - x1) * (y2 - y1);

  x_amount = spvals.spread_amount_x;
  y_amount = spvals.spread_amount_y;

  /* Initialize random stuff */
  srand (time (NULL));
  /* EL: This does not seem very useful. There are only 6 diffent angles:
     -3, -2, -1, 0, 1, 2 in radians. */
  angle_mod_value = G_PI * 2;
  angle_sub_value = angle_mod_value / 2;
  x_mod_value = x_amount + 1;
  x_sub_value = x_mod_value / 2;
  y_mod_value = y_amount + 1;
  y_sub_value = y_mod_value / 2;

  /* Spread the image.  This is done by going through every pixel
     in the source image and swapping it with some other random
     pixel.  The random pixel is located within an ellipse that is
     as high as the spread_amount_y parameter and as wide as the
     spread_amount_x parameter.  This is done by randomly selecting
     an angle and then multiplying the sine of the angle to a random
     number whose range is between -spread_amount_x/2 and spread_amount_x/2.
     The y coordinate is found by multiplying the cosine of the angle
     to the random value generated from spread_amount_y.  The reason
     that the spread is done this way is to make the end product more
     random looking.  To see a result of this, compare spreading a
     square with gimp 0.54 to spreading a square with this filter.
     The corners are less sharp with this algorithm.
   */

  /* Spread the image! */

  if (src_drawable == dest_drawable)
    {
      offset_x = 0;
      offset_y = 0;
    }
  else
    {
      offset_x = x1;
      offset_y = y1;
    }

  gimp_pixel_rgn_init (&dest_rgn, dest_drawable,
                       x1 - offset_x, y1 - offset_y, (x2 - x1), (y2 - y1),
                       TRUE, TRUE);

  for (pr = gimp_pixel_rgns_register (1, &dest_rgn);
       pr != NULL; pr = gimp_pixel_rgns_process (pr))
    {
      destrow = dest_rgn.data;

      for (y = dest_rgn.y; y < (dest_rgn.y + dest_rgn.h); y++)
        {
          dest = destrow;

          for (x = dest_rgn.x; x < (dest_rgn.x + dest_rgn.w); x++)
            {
              /* These are the pixel coordinates in src_drawable */
              const gint src_x = x + offset_x;
              const gint src_y = y + offset_y;

              /* get random angle, x distance, and y distance */
              xdist = (rand () % x_mod_value) - x_sub_value;
              ydist = (rand () % y_mod_value) - y_sub_value;
              angle = (rand () % angle_mod_value) - angle_sub_value;

              xi = src_x + floor (sin (angle) * xdist);
              yi = src_y + floor (cos (angle) * ydist);

              /* Only displace the pixel if it's within the bounds of the image. */
              if ((xi >= 0) && (xi < width) && (yi >= 0) && (yi < height))
                src_tile = spread_pixel (src_drawable, src_tile,
                                         xi, yi, &row, &col, pixel[0]);
              else
                {
                  /* Else just copy it */
                  src_tile = spread_pixel (src_drawable, src_tile,
                                           src_x, src_y,
                                           &row, &col, pixel[0]);
                }

              for (k = 0; k < bytes; k++)
                *dest++ = pixel[0][k];
            }                   /* for */

          destrow += dest_rgn.rowstride;;
        }                       /* for */
      progress += dest_rgn.w * dest_rgn.h;
      gimp_preview_progress_updater_set_fraction (progress_updater,
                                                  (double) progress /
                                                  (double) max_progress);
    }                           /* for  */

  if (src_tile)
    gimp_tile_unref (src_tile, FALSE);

  /*  update the region  */
  gimp_drawable_flush (dest_drawable);
  gimp_drawable_merge_shadow (dest_drawable->drawable_id, TRUE);
  gimp_drawable_update (dest_drawable->drawable_id, x1, y1, (x2 - x1),
                        (y2 - y1));
}                               /* spread */

static GtkWidget *
spread_settings (GtkWidget * parent, gpointer data)
{
  GimpDrawable *drawable = (GimpDrawable *) data;
  GtkWidget *size;
  GimpUnit unit;
  gdouble xres;
  gdouble yres;
  guint32 image_id;

  /*  Get the image resolution and unit  */
  image_id = gimp_drawable_image (drawable->drawable_id);
  gimp_image_get_resolution (image_id, &xres, &yres);
  unit = gimp_image_get_unit (image_id);

  /* sizeentries */
  size = gimp_coordinates_new (unit, "%a", TRUE, FALSE, 75,
                               GIMP_SIZE_ENTRY_UPDATE_SIZE,
                               /* When the chain button is active the coordinate widget is broken */
                               FALSE,
                               /* spvals.spread_amount_x == spvals.spread_amount_y, */
                               FALSE,
                               _("_Horizontal:"), spvals.spread_amount_x,
                               xres, 0, MAX (drawable->width,
                                             drawable->height), 0, 0,
                               _("_Vertical:"), spvals.spread_amount_y, yres,
                               0, MAX (drawable->width, drawable->height), 0,
                               0);
  g_signal_connect (G_OBJECT (size), "value-changed",
                    G_CALLBACK (spread_size_update_callback), NULL);

  gtk_container_set_border_width (GTK_CONTAINER (size), 4);

  pint.size = size;

  return size;
}

static void
spread_size_update_callback (GtkWidget * widget, gpointer data)
{
  spvals.spread_amount_x =
    gimp_size_entry_get_refval (GIMP_SIZE_ENTRY (pint.size), 0);

  spvals.spread_amount_y =
    gimp_size_entry_get_refval (GIMP_SIZE_ENTRY (pint.size), 1);

  preview_update_effect ();
}

static gint
spread_dialog (gint32 image_ID, GimpDrawable * drawable)
{
  EffectParam_t param = {
    _("Spread"),
    "spread",
    "filters/spread.html",
    spread_settings,
    drawable,
    spread_callback
  };

  return do_preview_dialog (drawable, &param);
}

static void
spread_callback (GtkWidget * widget, GimpPreviewUpdateEvent * event,
                 gpointer data)
{
  GimpPreview *preview = GIMP_PREVIEW (widget);
  GimpDrawable *drawable = (GimpDrawable *) data;
  GimpDrawable *dest_drawable;
  GimpPreviewProgressUpdater preview_progress_updater;

  dest_drawable =
    gimp_preview_get_temp_drawable (drawable, event->image_x, event->image_y,
                                    event->image_width, event->image_height);

  gimp_preview_progress_updater_init_for_preview (&preview_progress_updater,
                                                  preview);
  spread (drawable, dest_drawable, event->image_x, event->image_y,
          event->image_width, event->image_height, &preview_progress_updater);
  gimp_preview_draw_unscaled_drawable (preview, dest_drawable);
  gimp_preview_free_temp_drawable (dest_drawable);
}
